# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

class ODC2DeprecationWarning(DeprecationWarning):
    """
    Subclasss of Deprecation Warning for API elements that are deprecated in 1.9 and will be removed in 2.0.

    Provided to support suppression of 1.9 deprecation warnings in e.g. sandbox-like environments to prevent
    end-users from freaking out.
    """
    pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import pytest


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_cli_spatial_indexes(index, clirunner):
    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=True)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'create', 'epsg:3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0

    # Double creation succeeds silently
    runner = clirunner(['spindex', 'create', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0

    # Double creation succeeds silently
    runner = clirunner(['spindex', 'update', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0

    # Double creation succeeds silently
    runner = clirunner(['spindex', 'update', 'EPSG:3857'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=True)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'drop', '3577'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1
    runner = clirunner(['spindex', 'drop', '--force', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=True)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    # Drop non-existent spindex ignored.
    runner = clirunner(['spindex', 'drop', '--force', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_cli_spatial_index_create_and_update(index, clirunner):
    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=True)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" not in runner.output
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'create', '--update', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=True)
    assert "EPSG:4326" in runner.output
    assert "EPSG:3577" in runner.output
    assert runner.exit_code == 0

    runner = clirunner(['spindex', 'drop', '3577'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1
    runner = clirunner(['spindex', 'drop', '--force', '3577'], verbose_flag=False, expect_success=True)
    assert runner.exit_code == 0


@pytest.mark.parametrize('datacube_env_name', ('datacube',))
def test_cli_spatial_indexes_on_non_supporting_index(index, clirunner):
    runner = clirunner(['spindex', 'list'], verbose_flag=False, expect_success=False)
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'create', '3577'], verbose_flag=False, expect_success=False)
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'update', '3577'], verbose_flag=False, expect_success=False)
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'drop', '3577'], verbose_flag=False, expect_success=False)
    assert "does not support spatial indexes" in runner.output
    assert runner.exit_code == 1


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_cli_spatial_indexes_no_srids(index, clirunner):
    runner = clirunner(['spindex', 'create'], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'update'], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1

    runner = clirunner(['spindex', 'drop'], verbose_flag=False, expect_success=False)
    assert "Must supply at least one CRS" in runner.output
    assert runner.exit_code == 1


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_cli_spatial_indexes_bad_srid(index, clirunner):
    runner = clirunner(['spindex', 'create', '1'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1
    runner = clirunner(['spindex', 'create', '--update', '1'], verbose_flag=False, expect_success=False)
    assert "Skipping update" in runner.output
    assert runner.exit_code == 1
    runner = clirunner(['spindex', 'update', '1'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1
    runner = clirunner(['spindex', 'drop', '1'], verbose_flag=False, expect_success=False)
    assert runner.exit_code == 1



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import numpy as np

import pytest
from rasterio.enums import Resampling

from datacube.storage._read import (
    read_time_slice,
    read_time_slice_v2,
    pick_read_scale,
    rdr_geobox)

from datacube.testutils.io import RasterFileDataSource
from odc.geo.overlap import compute_reproject_roi
from odc.geo.roi import roi_shape, roi_is_empty
from odc.geo.geobox import GeoBox

from odc.geo import geobox as gbx

from datacube.testutils.io import (
    rio_slurp
)

from datacube.testutils.geom import (
    epsg3857,
    AlbersGS,
)


nearest_resampling_parametrize = pytest.mark.parametrize(
    "nearest_resampling", ['nearest', Resampling.nearest, Resampling.nearest.value]
)


def test_pick_read_scale():
    assert pick_read_scale(0.7) == 1
    assert pick_read_scale(1.3) == 1
    assert pick_read_scale(2.3) == 2
    assert pick_read_scale(1.99999) == 2


@nearest_resampling_parametrize
def test_read_paste(nearest_resampling, tmpdir):
    from datacube.testutils import mk_test_image
    from datacube.testutils.io import write_gtiff
    from pathlib import Path

    pp = Path(str(tmpdir))

    xx = mk_test_image(128, 64, nodata=None)
    assert (xx != -999).all()

    mm = write_gtiff(pp/'tst-read-paste-128x64-int16.tif', xx, nodata=None)

    def _read(geobox, resampling=nearest_resampling,
              fallback_nodata=-999,
              dst_nodata=-999,
              check_paste=False):
        with RasterFileDataSource(mm.path, 1, nodata=fallback_nodata).open() as rdr:
            if check_paste:
                # check that we are using paste
                rr = compute_reproject_roi(rdr_geobox(rdr), geobox)
                assert rr.paste_ok is True

            yy = np.full(geobox.shape, dst_nodata, dtype=rdr.dtype)
            roi = read_time_slice(rdr, yy, geobox, resampling, dst_nodata)
            return yy, roi

    # read native whole
    yy, roi = _read(mm.geobox)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # read native whole, no nodata case
    yy, roi = _read(mm.geobox, fallback_nodata=None)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # read native whole, ignoring small sub-pixel translation
    yy, roi = _read(gbx.translate_pix(mm.geobox, 0.3, -0.4), fallback_nodata=-33)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # no overlap between src and dst
    yy, roi = _read(gbx.translate_pix(mm.geobox, 10000, -10000))
    assert roi_is_empty(roi)

    # read with Y flipped
    yy, roi = _read(gbx.flipy(mm.geobox))
    np.testing.assert_array_equal(xx[::-1, :], yy)
    assert roi == np.s_[0:64, 0:128]

    # read with X flipped
    yy, roi = _read(gbx.flipx(mm.geobox))
    np.testing.assert_array_equal(xx[:, ::-1], yy)
    assert roi == np.s_[0:64, 0:128]

    # read with X and Y flipped
    yy, roi = _read(gbx.flipy(gbx.flipx(mm.geobox)))
    assert roi == np.s_[0:64, 0:128]
    np.testing.assert_array_equal(xx[::-1, ::-1], yy[roi])

    # dst is fully inside src
    sroi = np.s_[10:19, 31:47]
    yy, roi = _read(mm.geobox[sroi])
    np.testing.assert_array_equal(xx[sroi], yy[roi])

    # partial overlap
    yy, roi = _read(gbx.translate_pix(mm.geobox, -3, -10))
    assert roi == np.s_[10:64, 3:128]
    np.testing.assert_array_equal(xx[:-10, :-3], yy[roi])
    assert (yy[:10, :] == -999).all()
    assert (yy[:, :3] == -999).all()

    # scaling paste
    yy, roi = _read(gbx.zoom_out(mm.geobox, 2), check_paste=True)
    assert roi == np.s_[0:32, 0:64]
    np.testing.assert_array_equal(xx[1::2, 1::2], yy)


@nearest_resampling_parametrize
def test_read_with_reproject(nearest_resampling, tmpdir):
    from datacube.testutils import mk_test_image
    from datacube.testutils.io import write_gtiff
    from pathlib import Path

    pp = Path(str(tmpdir))

    xx = mk_test_image(128, 64, nodata=None)
    assert (xx != -999).all()
    tile = AlbersGS.tile_geobox((17, -40))[:64, :128]

    mm = write_gtiff(pp/'tst-read-with-reproject-128x64-int16.tif', xx,
                     crs=str(tile.crs),
                     resolution=tile.resolution.xy,
                     offset=tile.transform*(0, 0),
                     nodata=-999)
    assert mm.geobox == tile

    def _read(geobox,
              resampling=nearest_resampling,
              fallback_nodata=None,
              dst_nodata=-999):
        with RasterFileDataSource(mm.path, 1, nodata=fallback_nodata).open() as rdr:
            yy = np.full(geobox.shape, dst_nodata, dtype=rdr.dtype)
            roi = read_time_slice(rdr, yy, geobox, resampling, dst_nodata)
            return yy, roi

    geobox = gbx.pad(mm.geobox, 10)
    geobox = gbx.zoom_out(geobox, 0.873)
    yy, roi = _read(geobox)

    assert roi[0].start > 0 and roi[1].start > 0
    assert (yy[0] == -999).all()

    yy_expect, _ = rio_slurp(mm.path, geobox)
    np.testing.assert_array_equal(yy, yy_expect)

    geobox = gbx.zoom_out(mm.geobox[3:-3, 10:-10], 2.1)
    yy, roi = _read(geobox)

    assert roi_shape(roi) == geobox.shape
    assert not (yy == -999).any()

    geobox = GeoBox.from_geopolygon(mm.geobox.extent.to_crs(epsg3857).buffer(50),
                                    resolution=mm.geobox.resolution)

    assert geobox.extent.contains(mm.geobox.extent.to_crs(epsg3857))
    assert geobox.crs != mm.geobox.crs
    yy, roi = _read(geobox)
    assert roi[0].start > 0 and roi[1].start > 0
    assert (yy[0] == -999).all()

    geobox = gbx.zoom_out(geobox, 4)
    yy, roi = _read(geobox, resampling='average')
    nvalid = (yy != -999).sum()
    nempty = (yy == -999).sum()
    assert nvalid > nempty


@nearest_resampling_parametrize
def test_read_paste_v2(nearest_resampling, tmpdir):
    from datacube.testutils import mk_test_image
    from datacube.testutils.io import write_gtiff
    from datacube.testutils.iodriver import open_reader
    from pathlib import Path

    pp = Path(str(tmpdir))

    xx = mk_test_image(128, 64, nodata=None)
    assert (xx != -999).all()

    mm = write_gtiff(pp/'tst-read-paste-128x64-int16.tif', xx, nodata=None)

    def _read(geobox, resampling=nearest_resampling,
              fallback_nodata=-999,
              dst_nodata=-999,
              check_paste=False):

        rdr = open_reader(mm.path,
                          nodata=fallback_nodata)
        if check_paste:
            # check that we are using paste
            rr = compute_reproject_roi(rdr_geobox(rdr), geobox)
            assert rr.paste_ok is True

        yy = np.full(geobox.shape, dst_nodata, dtype=rdr.dtype)
        yy_, roi = read_time_slice_v2(rdr, geobox, resampling, dst_nodata)
        if yy_ is None:
            print(f"Got None out of read_time_slice_v2: {roi} must be empty")
        else:
            yy[roi] = yy_
        return yy, roi

    # read native whole
    yy, roi = _read(mm.geobox)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # read native whole, no nodata case
    yy, roi = _read(mm.geobox, fallback_nodata=None)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # read native whole, ignoring small sub-pixel translation
    yy, roi = _read(gbx.translate_pix(mm.geobox, 0.3, -0.4), fallback_nodata=-33)
    np.testing.assert_array_equal(xx, yy)
    assert roi == np.s_[0:64, 0:128]

    # no overlap between src and dst
    yy, roi = _read(gbx.translate_pix(mm.geobox, 10000, -10000))
    assert roi_is_empty(roi)

    # read with Y flipped
    yy, roi = _read(gbx.flipy(mm.geobox))
    np.testing.assert_array_equal(xx[::-1, :], yy)
    assert roi == np.s_[0:64, 0:128]

    # read with X flipped
    yy, roi = _read(gbx.flipx(mm.geobox))
    np.testing.assert_array_equal(xx[:, ::-1], yy)
    assert roi == np.s_[0:64, 0:128]

    # read with X and Y flipped
    yy, roi = _read(gbx.flipy(gbx.flipx(mm.geobox)))
    assert roi == np.s_[0:64, 0:128]
    np.testing.assert_array_equal(xx[::-1, ::-1], yy[roi])

    # dst is fully inside src
    sroi = np.s_[10:19, 31:47]
    yy, roi = _read(mm.geobox[sroi])
    np.testing.assert_array_equal(xx[sroi], yy[roi])

    # partial overlap
    yy, roi = _read(gbx.translate_pix(mm.geobox, -3, -10))
    assert roi == np.s_[10:64, 3:128]
    np.testing.assert_array_equal(xx[:-10, :-3], yy[roi])
    assert (yy[:10, :] == -999).all()
    assert (yy[:, :3] == -999).all()

    # scaling paste
    yy, roi = _read(gbx.zoom_out(mm.geobox, 2), check_paste=True)
    assert roi == np.s_[0:32, 0:64]
    np.testing.assert_array_equal(xx[1::2, 1::2], yy)


@nearest_resampling_parametrize
def test_read_with_reproject_v2(nearest_resampling, tmpdir):
    from datacube.testutils import mk_test_image
    from datacube.testutils.io import write_gtiff
    from datacube.testutils.iodriver import open_reader
    from pathlib import Path

    pp = Path(str(tmpdir))

    xx = mk_test_image(128, 64, nodata=None)
    assert (xx != -999).all()
    tile = AlbersGS.tile_geobox((17, -40))[:64, :128]

    def _read(geobox, resampling=nearest_resampling,
              fallback_nodata=-999,
              dst_nodata=-999):

        rdr = open_reader(mm.path,
                          nodata=fallback_nodata)

        yy = np.full(geobox.shape, dst_nodata, dtype=rdr.dtype)
        yy_, roi = read_time_slice_v2(rdr, geobox, resampling, dst_nodata)
        yy[roi] = yy_
        return yy, roi

    mm = write_gtiff(pp/'tst-read-with-reproject-128x64-int16.tif', xx,
                     crs=str(tile.crs),
                     resolution=tile.resolution.xy,
                     offset=tile.transform*(0, 0),
                     nodata=-999)
    assert mm.geobox == tile

    geobox = gbx.pad(mm.geobox, 10)
    geobox = gbx.zoom_out(geobox, 0.873)
    yy, roi = _read(geobox)

    assert roi[0].start > 0 and roi[1].start > 0
    assert (yy[0] == -999).all()

    yy_expect, _ = rio_slurp(mm.path, geobox)
    np.testing.assert_array_equal(yy, yy_expect)

    geobox = gbx.zoom_out(mm.geobox[3:-3, 10:-10], 2.1)
    yy, roi = _read(geobox)

    assert roi_shape(roi) == geobox.shape
    assert not (yy == -999).any()

    geobox = GeoBox.from_geopolygon(mm.geobox.extent.to_crs(epsg3857).buffer(50),
                                    resolution=mm.geobox.resolution)

    assert geobox.extent.contains(mm.geobox.extent.to_crs(epsg3857))
    assert geobox.crs != mm.geobox.crs
    yy, roi = _read(geobox)
    assert roi[0].start > 0 and roi[1].start > 0
    assert (yy[0] == -999).all()

    geobox = gbx.zoom_out(geobox, 4)
    yy, roi = _read(geobox, resampling='average')
    nvalid = (yy != -999).sum()
    nempty = (yy == -999).sum()
    assert nvalid > nempty



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import pytest
from datacube.model import Dataset
from datacube.testutils import suppress_deprecations


@pytest.mark.parametrize('datacube_env_name', ('datacube',))
def test_legacy_location_behaviour(index, ls8_eo3_dataset):
    with suppress_deprecations():
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == [ls8_eo3_dataset.uri]
        update = Dataset(  # Test of deprecated behaviour
            ls8_eo3_dataset.product,
            ls8_eo3_dataset.metadata_doc,
            uris=locations + ["file:/tmp/foo"])
        index.datasets.update(update)
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert index.datasets.get_location(ls8_eo3_dataset.id) == locations[0]
        assert locations == ["file:/tmp/foo", ls8_eo3_dataset.uri]
        index.datasets.add_location(ls8_eo3_dataset.id, "s3:/bucket/hole/straw.axe")
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == ["s3:/bucket/hole/straw.axe", "file:/tmp/foo", ls8_eo3_dataset.uri,]
        index.datasets.archive_location(ls8_eo3_dataset.id, "file:/tmp/foo")
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == ["s3:/bucket/hole/straw.axe", ls8_eo3_dataset.uri,]
        assert "file:/tmp/foo" in index.datasets.get_archived_locations(ls8_eo3_dataset.id)
        assert "file:/tmp/foo" == index.datasets.get_archived_location_times(ls8_eo3_dataset.id)[0][0]
        index.datasets.restore_location(ls8_eo3_dataset.id, "file:/tmp/foo")
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == ["s3:/bucket/hole/straw.axe", "file:/tmp/foo", ls8_eo3_dataset.uri,]
        index.datasets.remove_location(ls8_eo3_dataset.id, "file:/tmp/foo")
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == ["s3:/bucket/hole/straw.axe", ls8_eo3_dataset.uri,]
        index.datasets.remove_location(ls8_eo3_dataset.id, "s3:/bucket/hole/straw.axe")
        index.datasets.remove_location(ls8_eo3_dataset.id, ls8_eo3_dataset.uri)
        ls8_eo3_dataset = index.datasets.get(ls8_eo3_dataset.id)
        assert ls8_eo3_dataset.uri is None
        assert index.datasets.get_location(ls8_eo3_dataset.id) is None


@pytest.mark.parametrize('datacube_env_name', ('experimental',))
def test_postgis_no_multiple_locations(index, ls8_eo3_dataset):
    with suppress_deprecations():
        locations = index.datasets.get_locations(ls8_eo3_dataset.id)  # Test of deprecated method
        assert locations == [ls8_eo3_dataset.uri]

        update = Dataset(
            ls8_eo3_dataset.product,
            ls8_eo3_dataset.metadata_doc,
            uris=locations + ["file:/tmp/foo"])
        with pytest.raises(ValueError):
            index.datasets.update(update)
        assert index.datasets.get_location(ls8_eo3_dataset.id) == ls8_eo3_dataset.uri

        index.datasets.remove_location(ls8_eo3_dataset.id, "file:/tmp/foo")
        assert index.datasets.get_location(ls8_eo3_dataset.id) == ls8_eo3_dataset.uri

        index.datasets.remove_location(ls8_eo3_dataset.id, ls8_eo3_dataset.uri)
    ls8_eo3_dataset = index.datasets.get(ls8_eo3_dataset.id)
    assert ls8_eo3_dataset.uri is None
    assert index.datasets.get_location(ls8_eo3_dataset.id) is None

    with suppress_deprecations():
        index.datasets.add_location(ls8_eo3_dataset.id, "file:/tmp/foo")
        location = index.datasets.get_location(ls8_eo3_dataset.id)
        assert location == "file:/tmp/foo"

        with pytest.raises(ValueError):
            index.datasets.add_location(ls8_eo3_dataset.id, "s3:/bucket/hole/straw.axe")

        assert index.datasets.get_archived_locations(ls8_eo3_dataset.id) == []


def test_dataset_tuple_uris(ls8_eo3_product):
    from datacube.index.abstract import DatasetTuple
    dst1 = DatasetTuple(ls8_eo3_product, {"dummy": True}, "file:///uri1")
    dst2 = DatasetTuple(ls8_eo3_product, {"dummy": True}, ["file:///uri1", "https://host.domain/uri1"])

    with suppress_deprecations():
        assert dst1.uri == dst2.uri
        assert dst1.uri == dst2.uri
        assert dst1.uris == [dst1.uri]
        assert dst2.uri in dst2.uris
    assert not dst1.is_legacy
    assert dst2.is_legacy



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

from ._types import BatchStatus, DSID, DatasetTuple, dsid_to_uuid, DatasetSpatialMixin
from ._users import AbstractUserResource
from ._metadata_types import AbstractMetadataTypeResource, default_metadata_type_docs, _DEFAULT_METADATA_TYPES_PATH
from ._products import AbstractProductResource
from ._lineage import AbstractLineageResource, NoLineageResource
from ._datasets import AbstractDatasetResource
from ._transactions import AbstractTransaction, UnhandledTransaction
from ._index import AbstractIndex, AbstractIndexDriver

__all__ = [
    "BatchStatus",
    "DSID", "dsid_to_uuid",
    "DatasetTuple", "DatasetSpatialMixin",
    "default_metadata_type_docs", "_DEFAULT_METADATA_TYPES_PATH",
    "AbstractUserResource",
    "AbstractMetadataTypeResource", "AbstractProductResource",
    "AbstractLineageResource", "NoLineageResource",
    "AbstractDatasetResource",
    "AbstractTransaction", "UnhandledTransaction",
    "AbstractIndex", "AbstractIndexDriver",
]



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
"""
Storage Query and Access API module
"""


import logging
import datetime
import collections
import math
import warnings
from typing import Optional, Union
import pandas

from pandas import to_datetime as pandas_to_datetime
import numpy as np
from ..index import extract_geom_from_query, strip_all_spatial_fields_from_query
from ..model import Range, Dataset
from ..utils.dates import normalise_dt, tz_aware
from odc.geo import Geometry
from odc.geo.geom import lonlat_bounds, mid_longitude

_LOG = logging.getLogger(__name__)


class GroupBy:
    def __init__(self, group_by_func, dimension, units, sort_key=None, group_key=None):
        """
        GroupBy Object

        :param group_by_func: Dataset -> group identifier
        :param dimension: dimension of the group key
        :param units: units of the group key
        :param sort_key: how to sort datasets in a group internally
        :param group_key: the coordinate value for a group
                          list[Dataset] -> coord value
        """
        self.group_by_func = group_by_func

        self.dimension = dimension
        self.units = units

        if sort_key is None:
            sort_key = group_by_func
        self.sort_key = sort_key

        if group_key is None:
            group_key = lambda datasets: group_by_func(datasets[0])  # noqa: E731
        self.group_key = group_key


OTHER_KEYS = ('measurements', 'group_by', 'output_crs', 'resolution', 'set_nan', 'product', 'geopolygon', 'like',
              'source_filter')


class Query:
    def __init__(self, index=None, product=None, geopolygon=None, like=None, **search_terms):
        """Parses search terms in preparation for querying the Data Cube Index.

        Create a :class:`Query` object by passing it a set of search terms as keyword arguments.

        >>> query = Query(product='ls5_nbar_albers', time=('2001-01-01', '2002-01-01'))

        Use by accessing :attr:`search_terms`:

        >>> query.search_terms['time']  # doctest: +NORMALIZE_WHITESPACE
        Range(begin=datetime.datetime(2001, 1, 1, 0, 0, tzinfo=tzutc()), \
        end=datetime.datetime(2002, 1, 1, 23, 59, 59, 999999, tzinfo=tzutc()))

        By passing in an ``index``, the search parameters will be validated as existing on the ``product``,
        and a spatial search appropriate for the index driver can be extracted.

        Used by :meth:`datacube.Datacube.find_datasets` and :meth:`datacube.Datacube.load`.

        :param datacube.index.Index index: An optional `index` object, if checking of field names is desired.
        :param str product: name of product
        :type geopolygon: the spatial boundaries of the search, can be:
                          odc.geo.geom.Geometry: A Geometry object
                          Any string or JsonLike object that can be converted to a Geometry object.
                          An iterable of either of the above; or
                          None: no geopolygon defined (may be derived from like or lat/lon/x/y/crs search terms)
        :param xarray.Dataset like: spatio-temporal bounds of `like` are used for the search
        :param search_terms:
         * `measurements` - list of measurements to retrieve
         * `latitude`, `lat`, `y`, `longitude`, `lon`, `long`, `x` - tuples (min, max) bounding spatial dimensions
         * 'extra_dimension_name' (e.g. `z`) - tuples (min, max) bounding extra \
            dimensions specified by name for 3D datasets. E.g. z=(10, 30).
         * `crs` - spatial coordinate reference system to interpret the spatial bounds
         * `group_by` - observation grouping method. One of `time`, `solar_day`. Default is `time`
        """
        self.index = index
        self.product = product
        self.geopolygon = extract_geom_from_query(geopolygon=geopolygon, **search_terms)
        if 'source_filter' in search_terms and search_terms['source_filter'] is not None:
            self.source_filter = Query(**search_terms['source_filter'])
        else:
            self.source_filter = None

        search_terms = strip_all_spatial_fields_from_query(search_terms)
        remaining_keys = set(search_terms.keys()) - set(OTHER_KEYS)
        if self.index:
            # Retrieve known keys for extra dimensions
            known_dim_keys = set()
            if product is not None:
                datacube_products = index.products.search(product=product)
            else:
                datacube_products = index.products.get_all()

            for datacube_product in datacube_products:
                known_dim_keys.update(datacube_product.extra_dimensions.dims.keys())

            remaining_keys -= known_dim_keys

            unknown_keys = remaining_keys - set(index.products.get_field_names())
            # TODO: What about keys source filters, and what if the keys don't match up with this product...
            if unknown_keys:
                raise LookupError('Unknown arguments: ', unknown_keys)

        self.search = {}
        for key in remaining_keys:
            self.search.update(_values_to_search(**{key: search_terms[key]}))

        if like:
            assert self.geopolygon is None, "'like' with other spatial bounding parameters is not supported"
            self.geopolygon = getattr(like, 'extent', self.geopolygon)

            if 'time' not in self.search:
                time_coord = like.coords.get('time')
                if time_coord is not None:
                    self.search['time'] = _time_to_search_dims(
                        # convert from np.datetime64 to datetime.datetime
                        (pandas_to_datetime(time_coord.values[0]).to_pydatetime(),
                         pandas_to_datetime(time_coord.values[-1]).to_pydatetime())
                    )

    @property
    def search_terms(self):
        """
        Access the search terms as a dictionary.

        :type: dict
        """
        kwargs = {}
        kwargs.update(self.search)
        if self.geopolygon:
            if self.index and self.index.supports_spatial_indexes:
                kwargs['geopolygon'] = self.geopolygon
            else:
                geo_bb = lonlat_bounds(self.geopolygon, resolution="auto")
                if math.isclose(geo_bb.bottom, geo_bb.top, abs_tol=1e-5):
                    kwargs['lat'] = geo_bb.bottom
                else:
                    kwargs['lat'] = Range(geo_bb.bottom, geo_bb.top)
                if math.isclose(geo_bb.left, geo_bb.right, abs_tol=1e-5):
                    kwargs['lon'] = geo_bb.left
                else:
                    kwargs['lon'] = Range(geo_bb.left, geo_bb.right)
        if self.product:
            kwargs['product'] = self.product
        if self.source_filter:
            kwargs['source_filter'] = self.source_filter.search_terms
        return kwargs

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return """Datacube Query:
        type = {type}
        search = {search}
        geopolygon = {geopolygon}
        """.format(type=self.product,
                   search=self.search,
                   geopolygon=self.geopolygon)


def _extract_time_from_ds(ds: Dataset) -> datetime.datetime:
    return normalise_dt(ds.center_time)


def query_group_by(group_by='time', **kwargs):
    """
    Group by function for loading datasets

    :param group_by: group_by name, supported str are
    ::

        - time (default)
        - solar_day, see :func:`datacube.api.query.solar_day`

    or ::

        - :class:`datacube.api.query.GroupBy` object

    :return: :class:`datacube.api.query.GroupBy`
    :raises LookupError: when group_by string is not a valid dictionary key.
    """
    if isinstance(group_by, GroupBy):
        return group_by

    if not isinstance(group_by, str):
        group_by = None

    time_grouper = GroupBy(group_by_func=_extract_time_from_ds,
                           dimension='time',
                           units='seconds since 1970-01-01 00:00:00')

    solar_day_grouper = GroupBy(group_by_func=solar_day,
                                dimension='time',
                                units='seconds since 1970-01-01 00:00:00',
                                sort_key=_extract_time_from_ds,
                                group_key=lambda datasets: _extract_time_from_ds(datasets[0]))

    group_by_map = {
        None: time_grouper,
        'time': time_grouper,
        'solar_day': solar_day_grouper
    }

    try:
        return group_by_map[group_by]
    except KeyError:
        raise LookupError(
            f'No group by function for {group_by}, valid options are: {group_by_map.keys()}',
        )


def _value_to_range(value):
    if isinstance(value, (str, float, int)):
        value = float(value)
        return value, value
    else:
        return float(value[0]), float(value[-1])


def _values_to_search(**kwargs):
    search = {}
    for key, value in kwargs.items():
        if key.lower() in ('time', 't'):
            search['time'] = _time_to_search_dims(value)
        elif key not in ['latitude', 'lat', 'y'] + ['longitude', 'lon', 'x']:
            # If it's not a string, but is a sequence of length 2, then it's a Range
            if (
                not isinstance(value, str)
                and isinstance(value, collections.abc.Sequence)
                and len(value) == 2
            ):
                search[key] = Range(*value)
            # All other cases are default
            else:
                search[key] = value
    return search


def _time_to_search_dims(time_range):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        tr_start, tr_end = time_range, time_range

        if hasattr(time_range, '__iter__') and not isinstance(time_range, str):
            tmp = list(time_range)
            if len(tmp) > 2:
                raise ValueError("Please supply start and end date only for time query")

            tr_start, tr_end = tmp[0], tmp[-1]

        if isinstance(tr_start, (int, float)) or isinstance(tr_end, (int, float)):
            raise TypeError("Time dimension must be provided as a datetime or a string")

        if tr_start is None:
            start = datetime.datetime.fromtimestamp(0)
        elif not isinstance(tr_start, datetime.datetime):
            # convert to datetime.datetime
            if hasattr(tr_start, 'isoformat'):
                tr_start = tr_start.isoformat()
            start = pandas_to_datetime(tr_start).to_pydatetime()
        else:
            start = tr_start

        if tr_end is None:
            tr_end = datetime.datetime.now().strftime("%Y-%m-%d")
        # Attempt conversion to isoformat
        # allows pandas.Period to handle datetime objects
        if hasattr(tr_end, 'isoformat'):
            tr_end = tr_end.isoformat()
        # get end of period to ensure range is inclusive
        end = pandas.Period(tr_end).end_time.to_pydatetime()

        tr = Range(tz_aware(start), tz_aware(end))
        if start == end:
            return tr[0]

        return tr


def _convert_to_solar_time(utc, longitude):
    seconds_per_degree = 240
    offset_seconds = int(longitude * seconds_per_degree)
    offset = datetime.timedelta(seconds=offset_seconds)
    return utc + offset


def _ds_mid_longitude(dataset: Dataset) -> Optional[float]:
    m = dataset.metadata
    if hasattr(m, 'lon'):
        lon = m.lon
        return (lon.begin + lon.end)*0.5
    return None


def solar_day(dataset: Dataset, longitude: Optional[float] = None) -> np.datetime64:
    """
    Adjust Dataset timestamp for "local time" given location and convert to numpy.

    :param dataset: Dataset object from which to read time and location
    :param longitude: If supplied correct timestamp for this longitude,
                      rather than mid-point of the Dataset's footprint
    """
    utc = dataset.center_time.astimezone(datetime.timezone.utc)

    if longitude is None:
        _lon = _ds_mid_longitude(dataset)
        if _lon is None:
            raise ValueError('Cannot compute solar_day: dataset is missing spatial info')
        longitude = _lon

    solar_time = _convert_to_solar_time(utc, longitude)
    return np.datetime64(solar_time.date(), 'D')


def solar_offset(geom: Union[Geometry, Dataset],
                 precision: str = 'h') -> datetime.timedelta:
    """
    Given a geometry or a Dataset compute offset to add to UTC timestamp to get solar day right.

    This only work when geometry is "local enough".

    :param geom: Geometry with defined CRS
    :param precision: one of ``'h'`` or ``'s'``, defaults to hour precision
    """
    if isinstance(geom, Geometry):
        lon = mid_longitude(geom)
    else:
        _lon = _ds_mid_longitude(geom)
        if _lon is None:
            raise ValueError('Cannot compute solar offset, dataset is missing spatial info')
        lon = _lon

    if precision == 'h':
        return datetime.timedelta(hours=int(round(lon*24/360)))

    # 240 == (24*60*60)/360 (seconds of a day per degree of longitude)
    return datetime.timedelta(seconds=int(lon*240))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

"""
Datacube configuration
"""
import os
import warnings

from os import PathLike
from threading import Lock
from typing import Any, TypeAlias, Union, cast

from .cfg import find_config, parse_text
from .exceptions import ConfigException
from .opt import ODCOptionHandler, AliasOptionHandler, IndexDriverOptionHandler, BoolOptionHandler, IntOptionHandler
from .utils import ConfigDict, check_valid_env_name


# TypeAliases for more concise type hints
# (Unions required as typehint | operator doesn't work with string forward-references.
GeneralisedPath: TypeAlias = str | PathLike | list[str | PathLike]
GeneralisedCfg: TypeAlias = Union["ODCConfig", GeneralisedPath]
GeneralisedEnv: TypeAlias = Union["ODCEnvironment", str]
GeneralisedRawCfg: TypeAlias = str | ConfigDict


class ODCConfig:
    """
    Configuration finder/reader/parser.

    Attributes:
        allow_envvar_overrides: bool        If True, environment variables can override the values explicitly specified
                                            in the supplied configuration.

                                            Note that environments not explicitly specified in the supplied
                                            configuration (dynamic environments) can still be read from
                                            environment variables, even if this attribute is False.

        raw_text: str | None                The raw configuration text being used, as read from the configuration
                                            file or supplied directly by the user.  May be None if the user
                                            directly supplied configuration as a dictionary. May be in ini or yaml
                                            format.  Does not include dynamic environments or values overridden by
                                            environment variables.

        raw_config: dict[str, dict[str, Any]]   The raw dictionary form of the configuration, as supplied directly
                                                by the user, or as parsed from raw_text. Does not include dynamic
                                                environments or values overridden by environment variables.

        known_environments: dict[str, "ODCEnvironment"] A dictionary containing all environments defined in raw_config,
                                                        plus any dynamic environments read so far.
                                                        Environment themselves are not validated until read from.

        canonical_names: dict[str, list[str]]   A dictionary mapping canonical environment names to all aliases for
                                                that environment.
    """
    allow_envvar_overrides: bool = True
    raw_text: str | None = None
    raw_config: ConfigDict = {}
    known_environments: dict[str, "ODCEnvironment"] = {}
    canonical_names: dict[str, list[str]] = {}
    is_default = False

    def __init__(
            self,
            paths: GeneralisedPath | None = None,
            raw_dict: ConfigDict | None = None,
            text: str | None = None):
        """
        When called with no args, reads the first config file found in the config path list is used.
        The config path list is taken from:

        1) Environment variable $ODC_CONFIG_PATH (as a UNIX path style colon-separated path list)
        2) Environment variable $DATACUBE_CONFIG_PATH (as a UNIX path style colon-separated path list)
           This is a deprecated legacy environment variable, and please note that it's behaviour has changed
           slightly from datacube 1.8.x.
        3) The default config search path (i.e. .cfg._DEFAULT_CONFIG_SEARCH_PATH)

        If no config file is found at any of the paths in active path list, use the default configuration
        at , or if no such config file exists, use the default configuration (.cfg._DEFAULT_CONF). Configuration
        files may be in ini or yaml format. Environment variable overrides ARE applied.

        Otherwise, user may supply one (and only one) of the following:

        :param paths: The path of the configuration file, or a list of paths of candidate configuration files (the
                      first in the list that can be read is used).  If none of the supplied paths can be read, an
                      error is raised.  (Unlike calling with no arguments, the fallback default config is NOT
                      used.) Configuration file may be in ini or yaml format. Environment variable overrides ARE
                      applied.
        :param raw_dict: A raw dictionary containing configuration data.
                         Used as is - environment variable overrides are NOT applied.
        :param text: A string containing configuration data in ini or yaml format.
                     Used as is - environment variable overrides are NOT applied.
        """
        # Cannot supply both text AND paths.
        args_supplied: int = sum(map(lambda x: int(bool(x)), (paths, raw_dict, text)))
        if args_supplied > 1:
            raise ConfigException("Can only supply one of configuration path(s), raw dictionary, "
                                  "and explicit configuration text.")

        # Suppress environment variable overrides if explicit config text or dictionary is supplied.
        self.allow_envvar_overrides = not text and not raw_dict

        if not raw_dict and not text:
            # No explicit config passed in.  Check for ODC_CONFIG environment variables
            if os.environ.get("ODC_CONFIG"):
                text = os.environ["ODC_CONFIG"]
            else:
                # Read config text from config file
                text = find_config(paths, default_cb=self._set_default)

        self.raw_text = text
        if raw_dict is not None:
            self.raw_config = raw_dict
        else:
            self.raw_config = parse_text(cast(str, self.raw_text))

        self._aliases: dict[str, str] = {}
        self.known_environments: dict[str, ODCEnvironment] = {
            section: ODCEnvironment(self, section, self.raw_config[section], self.allow_envvar_overrides)
            for section in self.raw_config
        }
        self.canonical_names: dict[str, list[str]] = {}
        for alias, canonical in self._aliases.items():
            self.known_environments[alias] = self[canonical]
            if canonical in self.canonical_names:
                self.canonical_names[canonical].append(alias)
            else:
                self.canonical_names[canonical] = [canonical, alias]

    @classmethod
    def get_environment(cls,
                        env: GeneralisedEnv | None = None,
                        config: GeneralisedCfg | None = None,
                        raw_config: GeneralisedRawCfg | None = None) -> "ODCEnvironment":
        """
        Obtain an ODCConfig object from the most general possible arguments.

        It is an error to supply both config and raw_config, otherwise everything is optional and
        honours system defaults.

        :param env: An ODCEnvironment object or a string.
        :param config: An ODCConfig object or a config path.
        :param raw_config: A raw config string or ConfigDict.
        :return:
        """
        if config is not None and raw_config is not None:
            raise ConfigException("Cannot specify both config and raw_config")
        if isinstance(env, ODCEnvironment):
            return env
        else:
            if isinstance(config, ODCConfig):
                cfg = config
            elif isinstance(raw_config, str):
                cfg = ODCConfig(paths=config, text=raw_config)
            else:
                cfg = ODCConfig(paths=config, raw_dict=raw_config)
            return cfg[env]

    def _add_alias(self, alias: str, canonical: str) -> None:
        """
        Register an environment alias during ODCConfig class construction.

        Used internally by the Configuration library during class initialisation. Has no effect after initialisation.

        :param alias: The alias for the environment
        :param canonical: The canonical environment name the alias refers to
        """
        self._aliases[alias] = canonical

    def get_aliases(self, canonical_name: str) -> list[str]:
        """
        Get list of all possible names for a given canonical name.

        :param canonical_name: The canonical name of the target environment
        :return: A list of all known names for the target environment, including the canonical name itself and
                 any aliases.
        """
        if canonical_name in self.canonical_names:
            return self.canonical_names[canonical_name]
        else:
            return [canonical_name]

    def _set_default(self) -> None:
        self.is_default = True

    def __getitem__(self, item: str | None) -> "ODCEnvironment":
        """
        Environments can be accessed by name (canonical or aliases) with the getitem dunder method.

        E.g. cfg["env"]

        Passing in None returns the default environment, which is the first resolvable environment of:

        1. The environment named by the $ODC_ENVIRONMENT variable
        2. The environment named by the $DATACUBE_ENVIRONMENT variable
           (legacy environment variable name - now deprecated)
        3. The environment called `default` (dynamic environment lookup not supported)
        4. The environment called `datacube` (dynamic environment lookup not supported)

        If none of the above environment names are known, then a ConfigException is raised.

        If an explicit environment name is passed in that does not exist, dynamic environment lookup is attempted.
        Dynamic environment lookup always succeeds, but may return an environment which cannot connect to a database.

        :param item: A canonical environment name, an environment alias, or None.
        :return: A ODCEnvironment object
        """
        if item is None:
            # Get default.
            if os.environ.get("ODC_ENVIRONMENT"):
                item = os.environ["ODC_ENVIRONMENT"]
            elif os.environ.get("DATACUBE_ENVIRONMENT"):
                warnings.warn(
                    "Setting the default environment with $DATACUBE_ENVIRONMENT is deprecated. "
                    "Please use $ODC_ENVIRONMENT instead.")
                item = os.environ["DATACUBE_ENVIRONMENT"]
            elif "default" in self.known_environments:
                item = "default"
            elif "datacube" in self.known_environments:
                warnings.warn("Defaulting to the 'datacube' environment - "
                              "this fallback behaviour is deprecated and may change in a future release.")
                item = "datacube"
            else:
                # No explicitly defined (known) environments - assume default and hope there's config
                # available in environment variables.
                item = "default"
        if item not in self.known_environments:
            self.known_environments[item] = ODCEnvironment(self, item, {}, True)
        return self.known_environments[item]


class ODCEnvironment:
    """
    Configuration reader for an individual ODC environment.

    Only configuration options with a registered option handler are able to be read.  Configuration options
    may be read either as attributes on the ODCEnvironment objects or via the getitem dunder method.

    E.g.    env.index_driver   or    env["index_driver"]

    Attempting to access an unhandled or invalid option will raise a KeyError or AttributeError, as
    appropriate for the access method.

    ODCEnvironment objects should only be instantiated by and acquired from an ODCConfig object.
    """
    def __init__(self,
                 cfg: ODCConfig,
                 name: str,
                 raw: dict[str, Any],
                 allow_env_overrides: bool = True):
        self._cfg: ODCConfig = cfg
        check_valid_env_name(name)
        self._name: str = name
        self._raw: dict[str, Any] = raw
        self._allow_envvar_overrides: bool = allow_env_overrides
        self._lock = Lock()
        self._normalised: dict[str, Any] = {}

        if name == "user" and "default_environment" in raw:
            warnings.warn("The 'default_environment' setting in the 'user' section is no longer supported - "
                          "please refer to the documentation for more information")

        self._env_overrides_applied = False
        # Aliases are handled here, the alias OptionHandler is a place-holder.
        if "alias" in self._raw:
            alias = self._raw['alias']
            check_valid_env_name(alias)
            self._cfg._add_alias(self._name, alias)
            for opt in self._raw.keys():
                if opt != "alias":
                    raise ConfigException(
                        f"Alias environments should only contain an alias option. Extra option {opt} found.")

        self._option_handlers: list[ODCOptionHandler] = [
            AliasOptionHandler("alias", self),
            IndexDriverOptionHandler("index_driver", self, default="default"),
            BoolOptionHandler("skip_broken_datasets", self, default=False),
            IntOptionHandler("dc_load_limit", self, minval=0),
        ]

    def get_all_aliases(self):
        return self._cfg.get_aliases(self._name)

    def __getitem__(self, key: str) -> Any:
        with self._lock:
            if not self._normalised:
                # First access of environment - process config
                # Loop through content handlers.
                # Note that handlers may add more handlers to the end of the list while we are iterating over it.
                for handler in self._option_handlers:
                    self._handle_option(handler)
                if self._cfg.is_default and not self._env_overrides_applied:
                    warnings.warn("No configuration file found - using default configuration and environment variables")

        # Config already processed
        # 1. From Normalised
        if key in self._normalised:
            return self._normalised[key]
        # No config, no default.
        raise KeyError(key)

    def __getattr__(self, item: str) -> Any:
        try:
            return self[item]
        except KeyError:
            raise AttributeError(item)

    def _handle_option(self, handler: ODCOptionHandler) -> None:
        val = handler.get_val_from_environment()
        if val:
            self._env_overrides_applied = True
        else:
            val = self._raw.get(handler.name)
        val = handler.validate_and_normalise(val)
        self._normalised[handler.name] = val
        handler.handle_dependent_options(val)



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

import os
import warnings
from typing import Any, TYPE_CHECKING
from urllib.parse import urlparse
from urllib.parse import quote_plus

from .exceptions import ConfigException
from .utils import check_valid_option

if TYPE_CHECKING:
    from .api import ODCEnvironment


_DEFAULT_IAM_TIMEOUT = 600
_DEFAULT_CONN_TIMEOUT = 60
_DEFAULT_HOSTNAME = 'localhost'
_DEFAULT_DATABASE = 'datacube'

try:
    import pwd

    _DEFAULT_DB_USER: str | None = pwd.getpwuid(os.geteuid()).pw_name
except (ImportError, KeyError):
    # No default on Windows and some other systems
    _DEFAULT_DB_USER = None


class ODCOptionHandler:
    """
    Base option handler class.  Sufficient for basic option with no validation.

    Smarter or more specialised option handlers extend this class.
    """
    # If overridden by subclass to False, then no environment variable overrides apply.
    allow_envvar_lookup: bool = True

    def __init__(self, name: str, env: "ODCEnvironment", default: Any = None,
                 legacy_env_aliases=None):
        """

        :param name: Name of the option
        :param env: The ODCEnvironment the option is being read from
        :param default: The default value if not specified in the config file
        :param legacy_env_aliases:  Any legacy (pre odc-1.9) environment variable aliases for the option
        """
        check_valid_option(name)
        self.name: str = name
        self.env: "ODCEnvironment" = env
        self.default: Any = default
        if legacy_env_aliases:
            self.legacy_env_aliases = legacy_env_aliases
        else:
            self.legacy_env_aliases = []

    def validate_and_normalise(self, value: Any) -> Any:
        """
        Given a value read from a raw dictionary, return a normalised form.

        Subclasses should replace and call this implementation through super() for default handling.
        :param value: The value read from the raw dictionary (None if not present)
        :return: the normalised value.
        """
        if self.default is not None and value is None:
            return self.default
        return value

    def handle_dependent_options(self, value: Any) -> None:
        """
        Default implementation is no-op

        In subclasses:
        If the value of this option implies that dependent OptionHandlers should be run, they
        should be constructed here, and appended to self.env._option_handlers  (See examples below)

        :param value: The normalised value, as returned by validate_and_normalise
        :return: None
        """
        pass

    def get_val_from_environment(self) -> str | None:
        """
        Handle environment lookups.

        1. Returns None unless self.allow_envvar_lookups and self.env._allow_envvar_overrides are both set.
        2. Check canonical envvar name first.  E.g. option "bar" in environment "foo" -> $ODC_FOO_BAR
        3. Check canonical envvar name for any alias environments that point to this one.
        4. Check any legacy envvar names, and raise warnings if found.
        5. Check global envvar name, denoted by "all" instead of an environment name

        :return: First environment variable with non-empty value, or None if none found.
        """
        if self.allow_envvar_lookup and self.env._allow_envvar_overrides:
            canonical_name = f"odc_{self.env._name}_{self.name}".upper()
            for env_name in self.env.get_all_aliases():
                envvar_name = f"odc_{env_name}_{self.name}".upper()
                if val := os.environ.get(envvar_name):
                    return val
            for envvar_name in self.legacy_env_aliases:
                if val := os.environ.get(envvar_name):
                    warnings.warn(
                        f"Config being passed in by legacy environment variable ${envvar_name}. "
                        f"Please use ${canonical_name} instead.")
                    return val
            global_name = f"odc_all_{self.name}".upper()
            if val := os.environ.get(global_name):
                return val
        return None


class AliasOptionHandler(ODCOptionHandler):
    """
    Alias option is handled at the environment level.
    """
    allow_envvar_lookup: bool = False

    def validate_and_normalise(self, value: Any) -> Any:
        if value is None:
            return None
        raise ConfigException("Illegal attempt to directly access alias environment"
                              " - use the ODCConfig object to resolve the environment")


class IndexDriverOptionHandler(ODCOptionHandler):
    """
    Index Driver option.  Must be a valid index driver name, and index drivers may support further configuration.

    Example implementation for Postgresql/Postgis-based index drivers shown below.
    """
    def validate_and_normalise(self, value: Any) -> Any:
        value = super().validate_and_normalise(value)
        from datacube.drivers.indexes import index_drivers
        if value not in index_drivers():
            raise ConfigException(f"Unknown index driver: {value} - Try one of {','.join(index_drivers())}")
        return value

    def handle_dependent_options(self, value: Any) -> None:
        # Get driver-specific config options
        from datacube.drivers.indexes import index_driver_by_name
        driver = index_driver_by_name(value)
        assert driver is not None
        for option in driver.get_config_option_handlers(self.env):
            self.env._option_handlers.append(option)


class IntOptionHandler(ODCOptionHandler):
    """
    Require an integer value, with optional min and max vals.
    """
    def __init__(self, *args, minval: int | None = None, maxval: int | None = None, **kwargs):
        super().__init__(*args, **kwargs)
        self.minval = minval
        self.maxval = maxval

    def validate_and_normalise(self, value: Any) -> Any:
        # Call super() to get handle default value
        value = super().validate_and_normalise(value)
        if value is None:
            return value
        try:
            ival = int(value)
        except ValueError:
            raise ConfigException(f"Config option {self.name} must be an integer")
        if self.minval is not None and ival < self.minval:
            raise ConfigException(f"Config option {self.name} must be at least {self.minval}")
        if self.maxval is not None and ival > self.maxval:
            raise ConfigException(f"Config option {self.name} must not be greater than {self.minval}")
        return ival


class BoolOptionHandler(ODCOptionHandler):
    """
    Handle config option expecting a boolean value
    """
    def validate_and_normalise(self, value: Any) -> Any:
        value = super().validate_and_normalise(value)
        if isinstance(value, bool):
            return value
        elif isinstance(value, str) and value.lower() == "true":
            return True
        else:
            return False


class IAMAuthenticationOptionHandler(ODCOptionHandler):
    """
    A simple boolean, compatible with the historic behaviour of the IAM Authentication on/off option.

    y/yes (case-insensitive): True
    Anything else: False

    If true, adds an IAM Timeout Option to the Environment.
    """
    def validate_and_normalise(self, value: Any) -> Any:
        if isinstance(value, bool):
            return value
        elif isinstance(value, str) and value.lower() in ('y', 'yes'):
            return True
        else:
            return False

    def handle_dependent_options(self, value: Any) -> None:
        if value:
            self.env._option_handlers.append(
                IntOptionHandler("db_iam_timeout", self.env, default=_DEFAULT_IAM_TIMEOUT,
                                 legacy_env_aliases=['DATACUBE_IAM_TIMEOUT'],
                                 minval=1)
            )


class PostgresURLOptionHandler(ODCOptionHandler):
    def validate_and_normalise(self, value: Any) -> Any:
        if not value:
            return None
        components = urlparse(value)
        # Check URL scheme is postgresql:
        if components.scheme != "postgresql":
            raise ConfigException("Database URL is not a postgresql connection URL")
        # Don't bother splitting up the url, we'd just have to put it back together again later
        return value

    def handle_dependent_options(self, value: Any) -> None:
        if value is None:
            handlers: tuple[ODCOptionHandler, ...] = (
                ODCOptionHandler("db_username", self.env, legacy_env_aliases=['DB_USERNAME'],
                                 default=_DEFAULT_DB_USER),
                ODCOptionHandler("db_password", self.env, legacy_env_aliases=['DB_PASSWORD']),
                ODCOptionHandler("db_hostname", self.env, legacy_env_aliases=['DB_HOSTNAME'],
                                 default=_DEFAULT_HOSTNAME),
                IntOptionHandler("db_port", self.env, default=5432, legacy_env_aliases=['DB_PORT'],
                                 minval=1, maxval=49151),
                ODCOptionHandler("db_database", self.env, legacy_env_aliases=['DB_DATABASE'],
                                 default=_DEFAULT_DATABASE),
            )
        else:
            # These pseudo-handlers extract the equivalent config from the url returned by this handler.
            handlers = (
                PostgresURLPartHandler(self, "username", "db_username", self.env),
                PostgresURLPartHandler(self, "password", "db_password", self.env),
                PostgresURLPartHandler(self, "hostname", "db_hostname", self.env),
                PostgresURLPartHandler(self, "port", "db_port", self.env),
                PostgresURLPartHandler(self, "path", "db_database", self.env),
            )

        for handler in handlers:
            self.env._option_handlers.append(handler)


class PostgresURLPartHandler(ODCOptionHandler):
    def __init__(self, urlhandler: PostgresURLOptionHandler, urlpart: str, name: str, env: "ODCEnvironment"):
        self.urlhandler = urlhandler
        self.urlpart = urlpart
        super().__init__(name, env)

    def validate_and_normalise(self, value: Any) -> Any:
        url = self.env._normalised[self.urlhandler.name]
        purl = urlparse(url)
        part = getattr(purl, self.urlpart)
        if self.urlpart == "path" and part.startswith('/'):
            # Remove leading slash
            return str(part)[1:]
        else:
            return part

    def get_val_from_environment(self) -> str | None:
        # Never read from environment - take from URL, wherever it came from
        return None


def config_options_for_psql_driver(env: "ODCEnvironment"):
    """
       Config options for shared use by postgres-based index drivers
       (i.e. postgres and postgis drivers)
    """
    return [
        PostgresURLOptionHandler("db_url", env,
                                 legacy_env_aliases=['DATACUBE_DB_URL']),
        IAMAuthenticationOptionHandler("db_iam_authentication", env,
                                       legacy_env_aliases=['DATACUBE_IAM_AUTHENTICATION']),
        IntOptionHandler("db_connection_timeout", env, default=_DEFAULT_CONN_TIMEOUT, minval=1)
    ]


def psql_url_from_config(env: "ODCEnvironment"):
    if env.db_url:
        return env.db_url
    if not env.db_database:
        raise ConfigException(f"No database name supplied for environment {env._name}")
    url = "postgresql://"
    if env.db_username:
        if env.db_password:
            escaped_password = quote_plus(env.db_password)
            url += f"{env.db_username}:{escaped_password}@"
        else:
            url += f"{env.db_username}@"
    url += f"{env.db_hostname}:{env.db_port}/{env.db_database}"
    return url



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from typing import Tuple, Union, Optional, Any
from math import ceil

import numpy
import xarray as xr
import odc.geo.math as geomath
from odc.geo.xr import spatial_dims as xr_spatial_dims

from datacube.migration import ODC2DeprecationWarning
from deprecat import deprecat


def unsqueeze_data_array(da: xr.DataArray,
                         dim: str,
                         pos: int,
                         coord: Any = 0,
                         attrs: Optional[dict] = None) -> xr.DataArray:
    """
    Add a 1-length dimension to a data array.

    :param da: array to add a 1-length dimension
    :param dim: name of new dimension
    :param pos: position of dim
    :param coord: label of the coordinate on the unsqueezed dimension
    :param attrs: attributes for the coordinate dimension
    :return: A new xarray with a dimension added
    """
    new_dims = list(da.dims)
    new_dims.insert(pos, dim)
    new_shape = da.data.shape[:pos] + (1,) + da.data.shape[pos:]
    new_data = da.data.reshape(new_shape)
    new_coords = {k: v for k, v in da.coords.items()}
    new_coords[dim] = xr.DataArray([coord], dims=[dim], attrs=attrs)
    return xr.DataArray(new_data, dims=new_dims, coords=new_coords, attrs=da.attrs)


def unsqueeze_dataset(ds: xr.Dataset, dim: str, coord: int = 0, pos: int = 0) -> xr.Dataset:
    ds = ds.map(unsqueeze_data_array, dim=dim, pos=pos, keep_attrs=True, coord=coord)
    return ds


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def spatial_dims(xx: Union[xr.DataArray, xr.Dataset],
                 relaxed: bool = False) -> Optional[Tuple[str, str]]:
    return xr_spatial_dims(xx, relaxed)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def maybe_zero(x: float, tol: float) -> float:
    return geomath.maybe_zero(x, tol)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def maybe_int(x: float, tol: float) -> Union[int, float]:
    return geomath.maybe_int(x, tol)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def snap_scale(s, tol=1e-6):
    return geomath.snap_scale(s, tol)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def clamp(x, lo, up):
    return geomath.clamp(x, lo, up)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def is_almost_int(x: float, tol: float):
    return geomath.is_almost_int(x, tol)


def dtype_is_float(dtype) -> bool:
    """
    Check if `dtype` is floating-point.
    """
    return numpy.dtype(dtype).kind == 'f'


def valid_mask(xx, nodata):
    """
    Compute mask such that xx[mask] contains only valid pixels.
    """
    if dtype_is_float(xx.dtype):
        if nodata is None or numpy.isnan(nodata):
            return ~numpy.isnan(xx)
        return ~numpy.isnan(xx) & (xx != nodata)

    if nodata is None:
        return numpy.full_like(xx, True, dtype=bool)
    return xx != nodata


def invalid_mask(xx, nodata):
    """
    Compute mask such that xx[mask] contains only invalid pixels.
    """
    if dtype_is_float(xx.dtype):
        if nodata is None or numpy.isnan(nodata):
            return numpy.isnan(xx)
        return numpy.isnan(xx) | (xx == nodata)

    if nodata is None:
        return numpy.full_like(xx, False, dtype=bool)
    return xx == nodata


def num2numpy(x, dtype, ignore_range=None):
    """
    Cast python numeric value to numpy.

    :param x int|float: Numerical value to convert to numpy.type
    :param dtype str|numpy.dtype|numpy.type: Destination dtype
    :param ignore_range: If set to True skip range check and cast anyway (for example: -1 -> 255)
                         (Not supported in numpy 2.0+)

    :returns: None if x is None
    :returns: None if x is outside the valid range of dtype and ignore_range is not set
    :returns: dtype.type(x) if x is within range or ignore_range=True
    """
    if x is None:
        return None

    if isinstance(dtype, (str, type)):
        dtype = numpy.dtype(dtype)

    if ignore_range or dtype.kind == 'f':
        return dtype.type(x)

    info = numpy.iinfo(dtype)
    if info.min <= x <= info.max:
        return dtype.type(x)

    return None


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def data_resolution_and_offset(data, fallback_resolution=None):
    return geomath.data_resolution_and_offset(data, fallback_resolution)


@deprecat(reason='This method has been moved to odc-geo.', version='1.9.0', category=ODC2DeprecationWarning)
def affine_from_axis(xx, yy, fallback_resolution=None):
    return geomath.affine_from_axis(xx, yy, fallback_resolution)


def iter_slices(shape, chunk_size):
    """
    Generate slices for a given shape.

    E.g. ``shape=(4000, 4000), chunk_size=(500, 500)``
    Would yield 64 tuples of slices, each indexing 500x500.

    If the shape is not divisible by the chunk_size, the last chunk in each dimension will be smaller.

    :param tuple(int) shape: Shape of an array
    :param tuple(int) chunk_size: length of each slice for each dimension
    :return: Yields slices that can be used on an array of the given shape

    >>> list(iter_slices((5,), (2,)))
    [(slice(0, 2, None),), (slice(2, 4, None),), (slice(4, 5, None),)]
    """
    assert len(shape) == len(chunk_size)
    num_grid_chunks = [int(ceil(s / float(c))) for s, c in zip(shape, chunk_size)]
    for grid_index in numpy.ndindex(*num_grid_chunks):
        yield tuple(
            slice(min(d * c, stop), min((d + 1) * c, stop)) for d, c, stop in zip(grid_index, chunk_size, shape))



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
from dataclasses import dataclass
from enum import Enum
from uuid import UUID
from typing import Mapping, Optional, Sequence, Tuple, Iterable, Any, cast


class LineageDirection(Enum):
    """
    Enumeration specifying the direction sense of a LineageTree (source-ward or derived-ward)

     - SOURCES indicates a lineage tree that contains the source datasets of the root node
     - DERIVED indicates a lineage tree that contains the derived datasets of the root node
    """
    SOURCES = 1
    DERIVED = 2

    def opposite(self):
        if self == self.SOURCES:
            return self.DERIVED
        else:
            return self.SOURCES

    @property
    def label(self):
        if self == self.SOURCES:
            return "sources"
        else:
            return "derivations"


SerialisedTree = dict[str, str | dict[str, list["SerialisedTree"]]]


@dataclass
class LineageTree:
    """
    A node in a Dataset Lineage tree.

     - direction (LineageDirection): Whether this is a node in a source tree or a derived tree
     - dataset_id (UUID): The dataset id associated with this node
     - children (Optional[Mapping[str, Sequence[LineageTree]]]):
          An optional mapping of lineage nodes of the same direction as this node.
          The keys of the mapping are classifier strings.
          children=None means that there may be children in the database.
          children={} means there are no children in the database.
          children represent source datasets or derived datasets depending on the direction.
    home (Optional[str]):
          The home index associated with this node's dataset.
          Optional. Index drivers may not implement a home table, in which case this value
          will always be None.
    """
    direction: LineageDirection
    dataset_id: UUID
    children: dict[str, list["LineageTree"]] | None = None
    home: str | None = None

    def __eq__(self, other):
        if not self.children and not other.children:
            children_equal = True
        else:
            children_equal = (self.children == other.children)

        return (
            self.dataset_id == other.dataset_id
            and self.home == other.home
            and children_equal
        )

    def serialise(self, specify_direction_if_empty=True) -> SerialisedTree:
        serial: SerialisedTree = {"id": str(self.dataset_id)}
        if self.home:
            serial["home"] = self.home
        if self.children:
            serial[self.direction.label] = {
                classifier: [
                    child.serialise(False)
                    for child in children
                ]
                for classifier, children in self.children.items()
            }
        elif specify_direction_if_empty:
            serial[self.direction.label] = {}
        return serial

    @classmethod
    def deserialise(cls, serialised: SerialisedTree, direction: LineageDirection | None = None) -> "LineageTree":
        if "id" not in serialised:
            raise ValueError("Serialised Lineage tree node must have an id")
        id_ = UUID(cast(str, serialised["id"]))
        home = serialised.get("home")
        if direction is None:
            if LineageDirection.SOURCES.label in serialised:
                direction = LineageDirection.SOURCES
            elif LineageDirection.DERIVED.label in serialised:
                direction = LineageDirection.DERIVED
            else:
                raise ValueError(f"Ambiguous direction for serialised Lineage Tree at {id_}")
        if direction.opposite().label in serialised:
            raise ValueError(f"Ambiguous direction for serialised Lineage Tree at {id_}")
        if serialised.get(direction.label):
            children = {
                classifier: [
                    LineageTree.deserialise(child_tree, direction)
                    for child_tree in child_trees
                ]
                for classifier, child_trees in cast(
                    dict[str, list[SerialisedTree]],
                    serialised[direction.label]
                ).items()
            }
        else:
            children = {}
        return LineageTree(dataset_id=id_,
                           home=cast(str | None, home),
                           direction=direction, children=children)

    def find_subtree(self, dsid: UUID, _state: list["LineageTree"] | None = None) -> Optional["LineageTree"]:
        """
        Finds subtree with root at dsid, if there is one.

        Immediately returns the first match found with not-None children, but keeps track of first
        children=None match to return if no not-None matches.


        :param dsid: The desired dataset id
        :param _state: For shared state through recursion
        :return: None, or the subtree with root at dsid.
        """
        if _state is None:
            state: list[LineageTree] = []
            root_call = True
        else:
            state = _state
            root_call = False

        # root node will always be first match.
        if dsid == self.dataset_id:
            if self.children is not None:
                # Best match - shortcut
                return self
            if not state:
                # Partial match - cache if first
                state.append(self)
        # Recurse
        if self.children is not None:
            for classifier, children in self.children.items():
                for child in children:
                    subtree = child.find_subtree(dsid, _state=state)
                    if subtree:
                        # Catch and return shortcut best match
                        return subtree
        if root_call and state:
            # Cached first partial match
            return state[0]
        # No match
        return None

    @classmethod
    def from_eo3_doc(cls, doc: Mapping[str, Any],
                     home=None,
                     home_derived=None) -> "LineageTree":
        """
        Generate a shallow (depth=1) LineageTree from an EO3 dataset document

        :param doc: The dataset metadata dictionary
        :param home: Home database for source datasets (defaults to None).
        :param home_derived: Home database for the derived dataset (defaults to None).
        :return: A depth==1 LineageTree
        """
        lineage = doc.get("lineage")
        return cls.from_data(doc["id"], lineage, home=home, home_derived=home_derived)

    @classmethod
    def from_data(cls, dsid: UUID,
                  sources: Mapping[str, Sequence[UUID]] | None = None,
                  direction: LineageDirection = LineageDirection.SOURCES,
                  home: str | None = None,
                  home_derived: str | None = None) -> "LineageTree":
        """
        Generate a shallow (depth=1) LineageTree from the sort of data found in an EO3 dataset

        :param dsid: The (derived) dataset id
        :param sources: A dictionary of classifiers to list of source IDs
        :param direction: Tree direction (default SOURCEwards, as per an EO3 dataset)
        :param home: Home database for source datasets (defaults to None).
        :param home_derived: Home database for the derived dataset (defaults to None).
        :return: A depth==1 LineageTree
        """
        if not sources:
            children = None
        else:
            children = {
                classifier: [
                    cls(direction=direction, dataset_id=der, home=home)
                    for der in ders
                ]
                for classifier, ders in sources.items()
            }
        return cls(
            direction=direction,
            dataset_id=dsid,
            children=children,
            home=home_derived
        )

    def child_datasets(self) -> set[UUID]:
        child_dsids: set[UUID] = set()
        if self.children is None:
            return child_dsids
        for classifier, children in self.children.items():
            for child in children:
                subchildren = child.child_datasets()
                subchildren.add(child.dataset_id)
                if self.dataset_id in subchildren:
                    raise InconsistentLineageException("LineageTrees must be acyclic")
                child_dsids.update(subchildren)
        return child_dsids


class InconsistentLineageException(Exception):
    """
    Raised when a method would result in an inconsistent/invalid LineageTree or LineageRelations collection.
    """


@dataclass(frozen=True)
class LineageIDPair:
    """
    LineagePair

    A lineage relationship between a source UUID, a derived UUID
    (no classifier)
    """
    derived_id: UUID
    source_id: UUID


@dataclass(frozen=True)
class LineageRelation:
    """
    LineageRelation

    A lineage relationship between a source UUID, a derived UUID, with a classifier string
    """
    classifier: str
    source_id: UUID
    derived_id: UUID

    def ids(self):
        return LineageIDPair(derived_id=self.derived_id, source_id=self.source_id)


class LineageRelations:
    """
    An indexed collection of LineageRelations.

    For converting between iterables of LineageRelations and LineageTrees.
    Enforces all lineage chains are acyclic.
    """
    def __init__(self,
                 tree: LineageTree | None = None,
                 max_depth: int = 0,
                 relations: Iterable[LineageRelation] | None = None,
                 homes: Mapping[UUID, str] | None = None,
                 clone: Optional["LineageRelations"] = None) -> None:
        """
        All arguments are optional.  Default gives an empty LineageRelations, and:

             rels = LineageRelations(tree, max_depth=max_depth, relations=lrels, clone=clone)

        is equivalent to:

             rels = LineageRelations()
             rels.merge_tree(tree, max_depth=max_depth)
             rels.merge(clone)
             for rel in lrels:
                rels.merge_new_lineage_relation(rel)

        :param tree: Initially merge a LineageTree
        :param max_depth: The maximum depth to read the LineageTree.
                          Default/0: no limit.  Not used if tree is None.

        :param clone: Initially clone this other LineageRelations object
        """
        # Internal index of id/home relations
        self._homes: dict[UUID, str] = {}
        # Tuple[UUID, UUID]'s are always (derived, source)
        # Mapping  (derived, source): classifier - Allow search by source, derived pair.
        self._relations_idx: dict[LineageIDPair, str] = {}
        # Sequence of the distinct LineageRelation objects this object represents.
        self.relations: list[LineageRelation] = []
        # Mapping source to mapping derived to classifier.  Allow search by source
        self.by_source: dict[UUID, dict[UUID, str]] = {}
        # Mapping source to mapping derived to classifier.  Allow search by derived
        self.by_derived: dict[UUID, dict[UUID, str]] = {}
        # Dataset ids known to this object
        self.dataset_ids: set[UUID] = set()

        # Merge initial arguments
        if clone is not None:
            self.merge(clone)
        if tree is not None:
            self.merge_tree(tree, max_depth=max_depth)
        if relations is not None:
            for rel in relations:
                self.merge_new_lineage_relation(rel)
        if homes is not None:
            for id_, home in homes.items():
                self.merge_new_home(id_, home)

    def merge_new_home(self, id_: UUID, home: str) -> None:
        """
        Merge a new home relation

        Raises InconsistentLineageException if we already have this id with a different home

        :param id_: The dataset id
        :param home: The home string
        """
        if id_ in self._homes:
            if self._homes[id_] and self._homes[id_] != home:
                raise InconsistentLineageException(f"Tree contains inconsistent homes for dataset {id_}")
        else:
            self._homes[id_] = home

    def _merge_new_relation(self, ids: LineageIDPair, classifier: str) -> None:
        """
        Internal convenience wrapper to merge_new_lineage_relation
        """
        self.merge_new_lineage_relation(
            LineageRelation(
                classifier=classifier,
                derived_id=ids.derived_id,
                source_id=ids.source_id))

    def merge_new_lineage_relation(self, rel: LineageRelation) -> None:
        """
        Merge a new LineageRelation object

        Raises InconsistentLineageException if we already have this relation with a different classifier, or
        this relation would result in a cyclic relation.
        """
        ids = rel.ids()
        if ids in self._relations_idx:
            if self._relations_idx[ids] != rel.classifier:
                raise InconsistentLineageException(
                    f"Dataset {ids.derived_id} is derived from {ids.source_id} with inconsistent classifiers."
                )
        else:
            self._relations_idx[ids] = rel.classifier
            self.relations.append(rel)
            if rel.source_id not in self.by_source:
                self.by_source[rel.source_id] = {}
            if rel.derived_id not in self.by_derived:
                self.by_derived[rel.derived_id] = {}
            self.by_source[rel.source_id][rel.derived_id] = rel.classifier
            self.by_derived[rel.derived_id][rel.source_id] = rel.classifier
            # Check for cyclic dependencies:
            new_ids = set([ids.derived_id, ids.source_id])
            if new_ids & self.dataset_ids:
                # We already know about these ids so need to confirm we are still acyclic
                # Extract sourcewards from derived and vice versa for full tree coverage
                self.extract_tree(rel.derived_id, direction=LineageDirection.SOURCES)
                self.extract_tree(rel.source_id, direction=LineageDirection.DERIVED)
            self.dataset_ids.update(new_ids)

    def merge(self, pool: "LineageRelations") -> None:
        """
        Merge in another LineageRelations collection, ensuring it is consistent with this one.

        :param pool: The other LineageRelations object
        """
        for id_, home in pool._homes.items():
            self.merge_new_home(id_, home)
        for ids, classifier in pool._relations_idx.items():
            self._merge_new_relation(ids, classifier)

    def merge_tree(self, tree: LineageTree,
                   nodes: dict[UUID, LineageTree] | None = None,
                   max_depth: int = 0) -> None:
        """
        Merge in a LineageTree, ensuring it is consistent with the collection so far.

        Raises InconsistentLineageException if tree contains cyclic dependencies or inconsistent direction

        :param tree: The LineageTree to merge
        :param parent_node: The parent node (used to mark recursive traversal - should be None on first call)
        :param max_depth: The depth to traverse the tree to.  default/zero = unlimited
        """
        # Check new tree is acyclic within itself
        tree.child_datasets()
        if tree.home is not None:
            self.merge_new_home(tree.dataset_id, tree.home)
        # Determine recursion behaviour
        recurse = True
        next_max_depth = max_depth - 1
        if nodes is None:
            nodes = {}
        if max_depth == 0:
            next_max_depth = 0
        elif max_depth == 1:
            recurse = False
        if tree.children:
            if tree.dataset_id in nodes:
                raise InconsistentLineageException("Duplicate nodes in LineageTree")
            nodes[tree.dataset_id] = tree
        else:
            # tree.children is {} or None (i.e. leaf node of original input tree).
            # Try to extract a reverse-direction tree to check for cyclic dependencies
            self.extract_tree(tree.dataset_id, direction=tree.direction.opposite())
        if tree.children is None:
            return
        # Perform recursion, as determined above
        for classifier, children in tree.children.items():
            for child in children:
                if child.direction != tree.direction:
                    raise InconsistentLineageException("Tree contains both derived and source nodes")
                if tree.direction == LineageDirection.SOURCES:
                    ids = LineageIDPair(derived_id=tree.dataset_id, source_id=child.dataset_id)
                else:
                    ids = LineageIDPair(derived_id=child.dataset_id, source_id=tree.dataset_id)
                self._merge_new_relation(ids, classifier)
                if recurse:
                    self.merge_tree(
                        child,
                        nodes=nodes,
                        max_depth=next_max_depth
                    )
        return

    def relations_diff(self,
                       existing_relations: Optional["LineageRelations"] = None,
                       allow_updates: bool = False) -> Tuple[Mapping[LineageIDPair, str],
                                                             Mapping[LineageIDPair, str],
                                                             Mapping[UUID, str],
                                                             Mapping[UUID, str]]:
        """
        Compare to another LineageRelations object, returning records to be added to or updated in
        the other LinearRelations collection to consistently merge this collection into it.

        Intended to be used by index drivers when adding lineage data to an index.

        Raises InconsistentLineageException if updates are required and allow_updates is False, or if
        merging the two LineageRelations would result in cyclic dependencies.

        :param existing_relations: The relations currently in an index.
        :param allow_updates: Whether updates to existing records are allowed.
        :return: Tuple containing:
            Relations that need to be added to existing_relations to merge with this collection.
            Relations that need to be updated in existing_relations to merge with this collection.
            Homes that need to be added to existing_relations to merge with this collection.
            Homes that need to up updated in existing_relations to merge with this collection.
        """
        if not existing_relations:
            return (
                self._relations_idx, {},
                self._homes, {}
            )
        relations_to_add = {}
        relations_to_update = {}
        homes_to_add = {}
        homes_to_update = {}

        if not allow_updates:
            # Ensure no inconsistencies
            merged = LineageRelations(clone=self)
            merged.merge(existing_relations)
            # Determine homes and relations to add
            for id_, home in self._homes.items():
                if id_ not in existing_relations._homes:
                    homes_to_add[id_] = home
            for ids, classifier in self._relations_idx.items():
                if ids not in existing_relations._relations_idx:
                    relations_to_add[ids] = classifier
        else:
            # Determine homes to add and update
            for id_, home in self._homes.items():
                if id_ not in existing_relations._homes:
                    homes_to_add[id_] = home
                elif home != existing_relations._homes[id_]:
                    homes_to_update[id_] = home
            # Determine relations to add and update
            for ids, classifier in self._relations_idx.items():
                if ids not in existing_relations._relations_idx:
                    relations_to_add[ids] = classifier
                elif classifier != existing_relations._relations_idx[ids]:
                    relations_to_update[ids] = classifier
        return (
            relations_to_add, relations_to_update,
            homes_to_add, homes_to_update
        )

    def extract_tree(self,
                     root: UUID,
                     direction: LineageDirection = LineageDirection.SOURCES,
                     parents: set[UUID] | None = None,
                     so_far: set[UUID] | None = None,
                     ) -> LineageTree:
        """
        Extract a LineageTree from this LineageRelations collection.

        Used to detect cyclic dependencies.

        :param root: The dataset id at the root of the extracted LineageTree
        :param direction: The direction of the extracted tree
        :param parents: Used to detect cyclic dependencies in recursive mode
                        - should be None on initial call.
        :param so_far: Used to detect duplication from diamond dependencies in recursive mode
                       - should be None on initial call.
        :return: the extracted LineageTree.
        """
        # Trees are extracted from the root down, so the leaf-up cycle-detection of tree.child_datasets
        # is insufficient here
        if so_far is None:
            so_far = set()
        if parents is None:
            parents = set()
        if root in parents:
            raise InconsistentLineageException(f"LineageTrees must be acyclic: {root}")
        parents.add(root)
        if root in so_far:
            # Shortcut duplicates for handling diamond-dependencies
            return LineageTree(dataset_id=root, direction=direction, home=self._homes.get(root))
        so_far.add(root)

        children: dict[str, list[LineageTree]] = {}
        if direction == LineageDirection.SOURCES:
            subtrees = self.by_derived.get(root, {})
        else:
            subtrees = self.by_source.get(root, {})
        for dsid, classifier in subtrees.items():
            subtree = self.extract_tree(dsid, direction, set(parents), so_far)
            if classifier in children:
                children[classifier].append(subtree)
            else:
                children[classifier] = [subtree]
        tree = LineageTree(
            dataset_id=root,
            direction=direction,
            children=children,
            home=self._homes.get(root)
        )
        return tree



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0
import datetime

import pytest

from datacube.testutils import suppress_deprecations
from integration_tests.utils import ensure_datasets_are_indexed


# Current formulation of this test relies on non-EO3 test data
@pytest.mark.parametrize('datacube_env_name', ('datacube', ))
@pytest.mark.usefixtures('default_metadata_type',
                         'indexed_ls5_scene_products')
def test_index_get_product_time_bounds(index, clirunner, example_ls5_dataset_paths):
    def index_dataset(path):
        return clirunner(['dataset', 'add', str(path)])

    def index_products():
        valid_uuids = []
        for uuid, ls5_dataset_path in example_ls5_dataset_paths.items():
            valid_uuids.append(uuid)
            index_dataset(ls5_dataset_path)

        # Ensure that datasets are actually indexed
        ensure_datasets_are_indexed(index, valid_uuids)

        return valid_uuids

    valid_uuids = index_products()

    # lets get time values
    dataset_times = list(index.datasets.search_returning(field_names=('time',),
                                                         product='ls5_nbar_scene'))

    # get time bounds
    with suppress_deprecations():
        time_bounds = index.datasets.get_product_time_bounds(product='ls5_nbar_scene')  # Test of deprecated method
    left = sorted(dataset_times, key=lambda dataset: dataset.time.lower)[0].time.lower
    right = sorted(dataset_times, key=lambda dataset: dataset.time.upper)[-1].time.upper

    assert left == time_bounds[0]
    assert right == time_bounds[1]


def test_temporal_extent(
    index, ls8_eo3_dataset, ls8_eo3_dataset2, ls8_eo3_dataset3, ls8_eo3_dataset4
):
    with pytest.raises(KeyError):
        start, end = index.products.temporal_extent("orthentick_produckt")

    start, end = index.products.temporal_extent(ls8_eo3_dataset.product)
    assert start == datetime.datetime(
        2013, 4, 4, 0, 58, 34, 682275,
        tzinfo=datetime.timezone.utc)
    assert end == datetime.datetime(
        2016, 5, 28, 23, 50, 59, 149573,
        tzinfo=datetime.timezone.utc)
    start2, end2 = index.products.temporal_extent(ls8_eo3_dataset.product.name)
    assert start == start2 and end == end2
    try:
        start2, end2 = index.datasets.temporal_extent([
            ls8_eo3_dataset.id, ls8_eo3_dataset2.id,
            ls8_eo3_dataset3.id, ls8_eo3_dataset4.id,
        ])
        assert start == start2 and end == end2
    except NotImplementedError:
        # Not implemented by postgres driver
        pass



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0

# We often have one-arg-per column, so these checks aren't so useful.
# pylint: disable=too-many-arguments,too-many-public-methods,too-many-lines

# SQLAlchemy queries require "column == None", not "column is None" due to operator overloading:
# pylint: disable=singleton-comparison

"""
Persistence API implementation for postgis.
"""

import datetime
import json
import logging
import uuid  # noqa: F401
from sqlalchemy import (
    cast,
    delete,
    update,
    select,
    text,
    and_,
    or_,
    func,
    column,
)
from sqlalchemy.dialects.postgresql import insert
from sqlalchemy.sql.expression import Select
from sqlalchemy.dialects.postgresql import INTERVAL
from sqlalchemy.exc import IntegrityError

from typing import Iterable, Sequence, Optional, Set, Any
from typing import cast as type_cast

from datacube.index.fields import OrExpression
from datacube.model import Range
from odc.geo import CRS, Geometry
from datacube.utils.uris import split_uri
from datacube.index.abstract import DSID
from datacube.model.lineage import LineageRelation, LineageDirection
from . import _core
from ._fields import parse_fields, Expression, PgField, PgExpression, DateRangeDocField  # noqa: F401
from ._fields import NativeField, DateDocField, SimpleDocField, UnindexableValue
from ._schema import MetadataType, Product, Dataset, DatasetLineage, \
    search_field_index_map, search_field_indexes, DatasetHome
from ._spatial import geom_alchemy, generate_dataset_spatial_values, extract_geometry_from_eo3_projection
from .sql import escape_pg_identifier
from ...utils.changes import Offset

_LOG = logging.getLogger(__name__)


# Make a function because it's broken
def _dataset_select_fields() -> tuple:
    return tuple(f.alchemy_expression for f in _dataset_fields())


def _base_known_fields():
    fields = get_native_fields().copy()
    fields['archived'] = NativeField(
        'archived',
        'Archived date',
        Dataset.archived
    )
    return fields


def _dataset_fields() -> tuple:
    native_flds = get_native_fields()
    return (
        native_flds["id"],
        native_flds["indexed_time"],
        native_flds["indexed_by"],
        native_flds["product_id"],
        native_flds["metadata_type_id"],
        native_flds["metadata_doc"],
        NativeField(
            'archived',
            'Archived date',
            Dataset.archived
        ),
        native_flds["uri"]
    )


def _dataset_bulk_select_fields() -> tuple:
    return (
        Dataset.product_ref,
        Dataset.metadata_doc,
        Dataset.uri
    )


def get_native_fields() -> dict[str, NativeField]:
    # Native fields (hard-coded into the schema)
    fields = {
        'id': NativeField(
            'id',
            'Dataset UUID',
            Dataset.id
        ),
        'indexed_time': NativeField(
            'indexed_time',
            'When dataset was indexed',
            Dataset.added
        ),
        'indexed_by': NativeField(
            'indexed_by',
            'User who indexed the dataset',
            Dataset.added_by
        ),
        'product': NativeField(
            'product',
            'Product name',
            Product.name,
            join_clause=(Product.id == Dataset.product_ref)
        ),
        'product_id': NativeField(
            'product_id',
            'ID of a dataset type',
            Dataset.product_ref
        ),
        'metadata_type': NativeField(
            'metadata_type',
            'Metadata type name of dataset',
            MetadataType.name,
            join_clause=(MetadataType.id == Dataset.metadata_type_ref),
        ),
        'metadata_type_id': NativeField(
            'metadata_type_id',
            'ID of a metadata type',
            Dataset.metadata_type_ref
        ),
        'metadata_doc': NativeField(
            'metadata_doc',
            'Full metadata document',
            Dataset.metadata_doc
        ),
        'uri': NativeField(
            'uri',
            "Dataset URI",
            Dataset.uri_body,
            alchemy_expression=Dataset.uri,
        ),
    }
    return fields


def mk_simple_offset_field(field_name: str, description: str, offset: Offset) -> SimpleDocField:
    return SimpleDocField(
        name=field_name, description=description,
        alchemy_column=Dataset.metadata_doc,
        indexed=False,
        offset=offset
    )


def get_dataset_fields(metadata_type_definition):
    dataset_section = metadata_type_definition['dataset']

    fields = get_native_fields()
    # "Fixed fields" (not dynamic: defined in metadata type schema)
    fields.update(dict(
        creation_time=DateDocField(
            'creation_time',
            'Time when dataset was created (processed)',
            Dataset.metadata_doc,
            False,
            offset=dataset_section.get('creation_dt') or ['creation_dt']
        ),
        format=mk_simple_offset_field(
            'format', 'File format (GeoTiff, NetCDF)',
            dataset_section.get('format') or ['format', 'name']
        ),
        label=mk_simple_offset_field(
            'label', 'Label',
            dataset_section.get('label') or ['label']
        ),
    ))

    # noinspection PyTypeChecker
    fields.update(
        parse_fields(
            dataset_section['search_fields'],
            Dataset.metadata_doc
        )
    )
    return fields


def non_native_fields(mdt_metadata):
    return {
        name: field
        for name, field in get_dataset_fields(mdt_metadata).items()
        if not isinstance(field, NativeField)
    }


def extract_dataset_search_fields(ds_metadata, mdt_metadata):
    """
    :param ds_metdata: A Dataset metadata document
    :param mdt_metadata: The corresponding metadata-type definition document

    :return: A dictionary mapping search field names to (type_name, value) tuples.
    """
    return extract_dataset_fields(ds_metadata, non_native_fields(mdt_metadata))


def extract_dataset_fields(ds_metadata, fields):
    """
    :param ds_metdata: A Dataset metadata document
    :param fields: A dictionary of field names to Field objects

    :return: A dictionary mapping search field names to (type_name, value) tuples.
    """
    result = {}
    for field_name, field in fields.items():
        try:
            fld_type = field.type_name
            raw_val = field.extract(ds_metadata)
            sqla_val = field.search_value_to_alchemy(raw_val)
            result[field_name] = (fld_type, sqla_val)
        except UnindexableValue:
            continue
    return result


# Min/Max aggregating time fields for temporal_extent methods
time_min = DateDocField('acquisition_time_min',
                        'Min of time when dataset was acquired',
                        Dataset.metadata_doc,
                        False,  # is it indexed
                        offset=[
                            ['properties', 'dtr:start_datetime'],
                            ['properties', 'datetime']
                        ],
                        selection='least')


time_max = DateDocField('acquisition_time_max',
                        'Max of time when dataset was acquired',
                        Dataset.metadata_doc,
                        False,  # is it indexed
                        offset=[
                            ['properties', 'dtr:end_datetime'],
                            ['properties', 'datetime']
                        ],
                        selection='greatest')


class PostgisDbAPI:
    def __init__(self, parentdb, connection):
        self._db = parentdb
        self._connection = connection
        self._sqla_txn = None

    @property
    def in_transaction(self):
        return self._connection.in_transaction()

    def begin(self):
        self._connection.execution_options(isolation_level="REPEATABLE READ")
        self._sqla_txn = self._connection.begin()

    def _end_transaction(self):
        self._sqla_txn = None
        self._connection.execution_options(isolation_level="AUTOCOMMIT")

    def commit(self):
        self._sqla_txn.commit()
        self._end_transaction()

    def rollback(self):
        self._sqla_txn.rollback()
        self._end_transaction()

    def execute(self, command):
        return self._connection.execute(command)

    def insert_dataset(self, metadata_doc, dataset_id, product_id):
        """
        Insert dataset if not already indexed.
        :type metadata_doc: dict
        :type dataset_id: str or uuid.UUID
        :type product_id: int
        :return: whether it was inserted
        :rtype: bool
        """
        metadata_subquery = select(Product.metadata_type_ref).where(Product.id == product_id).scalar_subquery()
        ret = self._connection.execute(
            insert(Dataset).values(
                id=dataset_id,
                product_ref=product_id,
                metadata=metadata_doc,
                metadata_type_ref=metadata_subquery
            ).on_conflict_do_nothing(
                index_elements=['id']
            )
        )
        return ret.rowcount > 0

    def insert_dataset_bulk(self, values):
        requested = len(values)
        res = self._connection.execute(
            insert(Dataset), values
        )
        return res.rowcount, requested - res.rowcount

    def update_dataset(self, metadata_doc, dataset_id, product_id):
        """
        Update dataset
        :type metadata_doc: dict
        :type dataset_id: str or uuid.UUID
        :type product_id: int
        """
        res = self._connection.execute(
            update(Dataset).returning(Dataset.id).where(
                Dataset.id == dataset_id
            ).where(
                Dataset.product_ref == product_id
            ).values(
                metadata=metadata_doc
            )
        )
        return res.rowcount > 0

    def insert_dataset_location(self, dataset_id, uri):
        """
        Add a location to a dataset if it is not already recorded.

        Returns True if success, False if this location already existed

        :type dataset_id: str or uuid.UUID
        :type uri: str
        :rtype bool:
        """

        scheme, body = split_uri(uri)

        r = self._connection.execute(
            update(Dataset).returning(Dataset.uri).where(
                Dataset.id == dataset_id
            ).values(
                uri_scheme=scheme,
                uri_body=body
            )
        )

        return r.rowcount > 0

    def insert_dataset_search(self, search_table, dataset_id, key, value):
        """
        Add/update a search field index entry for a dataset

        Returns True on success

        :type search_table: A DatasetSearch ORM table
        :type dataset_id: str or uuid.UUID
        :type key: The name of the search field
        :type value: The value for the search field for this dataset.
        :rtype bool:
        """
        if isinstance(value, Range):
            value = list(value)
        r = self._connection.execute(
            insert(
                search_table
            ).values(
                dataset_ref=dataset_id,
                search_key=key,
                search_val=value,
            ).on_conflict_do_update(
                index_elements=[search_table.dataset_ref, search_table.search_key],
                set_=dict(search_val=value)
            )
        )
        return r.rowcount > 0

    def insert_dataset_search_bulk(self, search_type, values):
        search_table = search_field_index_map[search_type]
        r = self._connection.execute(insert(search_table).values(values))
        return r.rowcount

    def insert_dataset_spatial(self, dataset_id, crs, extent):
        """
        Add/update a spatial index entry for a dataset

        Returns True on success

        :type dataset_id: str or uuid.UUID
        :type crs: CRS
        :type extent: Geometry
        :rtype bool:
        """
        values = generate_dataset_spatial_values(dataset_id, crs, extent)
        if values is None:
            return False
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        r = self._connection.execute(
            insert(
                SpatialIndex
            ).values(
                **values
            ).on_conflict_do_update(
                index_elements=[SpatialIndex.dataset_ref],
                set_=dict(extent=values["extent"])
            )
        )
        return r.rowcount > 0

    def insert_dataset_spatial_bulk(self, crs, values):
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        r = self._connection.execute(insert(SpatialIndex).values(values))
        return r.rowcount

    def spatial_extent(self, ids, crs):
        SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
        if SpatialIndex is None:
            # Requested a CRS that has no spatial index, so use 4326 (which always has a spatial index)
            # and reproject to requested CRS.
            return self.spatial_extent(ids, CRS("epsg:4326")).to_crs(crs)
        query = select(
            func.ST_AsGeoJSON(func.ST_Union(SpatialIndex.extent))
        ).select_from(
            SpatialIndex
        ).where(
            SpatialIndex.dataset_ref.in_(ids)
        )
        result = self._connection.execute(query)
        for r in result:
            extent_json = r[0]
            if extent_json is None:
                return None
            return Geometry(json.loads(extent_json), crs=crs)
        return None

    def contains_dataset(self, dataset_id):
        return bool(
            self._connection.execute(
                select(Dataset.id).where(
                    Dataset.id == dataset_id
                )
            ).fetchone()
        )

    def datasets_intersection(self, dataset_ids):
        """ Compute set intersection: db_dataset_ids & dataset_ids
        """
        return [ds.id for ds in self._connection.execute(
                select(
                    Dataset.id
                ).where(
                    Dataset.id.in_(dataset_ids)
                )
            ).fetchall()
        ]

    def get_datasets_for_location(self, uri, mode=None):
        scheme, body = split_uri(uri)

        if mode is None:
            mode = 'exact' if body.count('#') > 0 else 'prefix'

        if mode == 'exact':
            body_query = Dataset.uri_body == body
        elif mode == 'prefix':
            body_query = Dataset.uri_body.startswith(body)
        else:
            raise ValueError('Unsupported query mode {}'.format(mode))

        return self._connection.execute(
            select(
                *_dataset_select_fields()
            ).where(
                and_(Dataset.uri_scheme == scheme, body_query)
            )
        ).fetchall()

    def all_dataset_ids(self, archived: bool | None = False):
        query = select(Dataset.id)
        if archived:
            query = query.where(
                Dataset.archived.is_not(None)
            )
        elif archived is not None:
            query = query.where(
                Dataset.archived.is_(None)
            )
        return self._connection.execute(query).fetchall()

    def archive_dataset(self, dataset_id):
        r = self._connection.execute(
            update(Dataset).where(
                Dataset.id == dataset_id
            ).where(
                Dataset.archived == None
            ).values(
                archived=func.now()
            )
        )
        return r.rowcount > 0

    def restore_dataset(self, dataset_id):
        r = self._connection.execute(
            update(Dataset).where(
                Dataset.id == dataset_id
            ).values(
                archived=None
            )
        )
        return r.rowcount > 0

    def delete_dataset(self, dataset_id):
        for table in search_field_indexes.values():
            self._connection.execute(
                delete(table).where(table.dataset_ref == dataset_id)
            )
        for crs in self._db.spatially_indexed_crses():
            SpatialIndex = self._db.spatial_index(crs)  # noqa: N806
            self._connection.execute(
                delete(
                    SpatialIndex
                ).where(
                    SpatialIndex.dataset_ref == dataset_id
                )
            )
        r = self._connection.execute(
            delete(Dataset).where(
                Dataset.id == dataset_id
            )
        )
        return r.rowcount > 0

    def get_dataset(self, dataset_id):
        return self._connection.execute(
            select(*_dataset_select_fields()).where(Dataset.id == dataset_id)
        ).first()

    def get_datasets(self, dataset_ids):
        return self._connection.execute(
            select(*_dataset_select_fields()).where(Dataset.id.in_(dataset_ids))
        ).fetchall()

    def get_derived_datasets(self, dataset_id):
        raise NotImplementedError()

    def get_dataset_sources(self, dataset_id):
        raise NotImplementedError()

    def search_datasets_by_metadata(self, metadata, archived):
        """
        Find any datasets that have the given metadata.

        :type metadata: dict
        :rtype: dict
        """
        # Find any storage types whose 'dataset_metadata' document is a subset of the metadata.
        where = Dataset.metadata_doc.contains(metadata)
        if archived:
            where = and_(where, Dataset.archived.is_not(None))
        elif archived is not None:
            where = and_(where, Dataset.archived.is_(None))
        query = select(*_dataset_select_fields()).where(where)
        return self._connection.execute(query).fetchall()

    def search_products_by_metadata(self, metadata):
        """
        Find any datasets that have the given metadata.

        :type metadata: dict
        :rtype: dict
        """
        # Find any storage types whose 'dataset_metadata' document is a subset of the metadata.
        return self._connection.execute(
            select(Product).where(Product.metadata_doc.contains(metadata))
        ).fetchall()

    @staticmethod
    def _alchemify_expressions(expressions):
        def raw_expr(expression):
            if isinstance(expression, OrExpression):
                return or_(raw_expr(expr) for expr in expression.exprs)
            return expression.alchemy_expression

        return [raw_expr(expression) for expression in expressions]

    def geospatial_query(self, geom):
        if not geom.crs:
            raise ValueError("Search geometry must have a CRS")
        SpatialIndex = self._db.spatial_index(geom.crs)  # noqa: N806
        if SpatialIndex is None:
            _LOG.info("No spatial index for crs %s - converting to 4326", geom.crs)
            default_crs = CRS("EPSG:4326")
            geom = geom.to_crs(default_crs)
            SpatialIndex = self._db.spatial_index(default_crs)  # noqa: N806
        geom_sql = geom_alchemy(geom)
        _LOG.info("query geometry = %s (%s)", geom.json, geom.crs)
        spatialquery = func.ST_Intersects(SpatialIndex.extent, geom_sql)
        return SpatialIndex, spatialquery

    def search_datasets_query(self,
                              expressions, source_exprs=None,
                              select_fields=None, with_source_ids=False, limit=None, geom=None,
                              archived: bool | None = False, order_by=None):
        """
        :type expressions: Tuple[Expression]
        :type source_exprs: Tuple[Expression]
        :type select_fields: Iterable[PgField]
        :type with_source_ids: bool
        :type limit: int
        :type order_by: Iterable[Any]
        :type geom: Geometry
        :rtype: sqlalchemy.Expression
        """
        # TODO: lineage handling and source search
        assert source_exprs is None
        assert not with_source_ids

        if not select_fields:
            select_fields = _dataset_fields()

        known_fields = _base_known_fields() | {f.name: f for f in select_fields}

        def _ob_exprs(o):
            if isinstance(o, str):
                if known_fields.get(o.lower()) is not None:
                    return known_fields[o.lower()].alchemy_expression
                raise ValueError(f"Cannot order by unknown field {o}")
            elif isinstance(o, PgField):
                return o.alchemy_expression
            else:
                # assume func, clause, or other expression, and leave as-is
                return o

        if order_by is not None:
            order_by = [_ob_exprs(o) for o in order_by]
        else:
            order_by = []

        select_columns = tuple(
            f.alchemy_expression.label(f.name)
            for f in select_fields
        )
        if geom:
            SpatialIndex, spatialquery = self.geospatial_query(geom)
        else:
            spatialquery = None
            SpatialIndex = None  # noqa: N806

        raw_expressions = PostgisDbAPI._alchemify_expressions(expressions)
        join_tables = PostgisDbAPI._join_tables(expressions, select_fields)
        if archived:
            # True: Return archived datasets ONLY
            where_expr = and_(Dataset.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            # False: Return active datasets ONLY
            where_expr = and_(Dataset.archived.is_(None), *raw_expressions)
        else:
            # None: Return BOTH active and archived datasets
            where_expr = and_(*raw_expressions)
        query = select(*select_columns).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)
        if spatialquery is not None:
            where_expr = and_(where_expr, spatialquery)
            query = query.join(SpatialIndex)
        query = query.where(where_expr).order_by(*order_by).limit(limit)
        return query

    def search_datasets(self, expressions,
                        source_exprs=None, select_fields=None,
                        with_source_ids=False, limit=None, geom=None,
                        archived: bool | None = False, order_by=None):
        """
        :type with_source_ids: bool
        :type select_fields: tuple[datacube.drivers.postgis._fields.PgField]
        :type expressions: tuple[datacube.drivers.postgis._fields.PgExpression]

        :return: An iterable of tuples of decoded values
        """
        assert source_exprs is None
        assert not with_source_ids
        if select_fields is None:
            select_fields = _dataset_fields()
        select_query = self.search_datasets_query(expressions, select_fields=select_fields, limit=limit,
                                                  geom=geom, archived=archived, order_by=order_by)
        _LOG.debug("search_datasets SQL: %s", str(select_query))

        def decode_row(raw: Iterable[Any]) -> dict[str, Any]:
            return {f.name: f.normalise_value(r) for r, f in zip(raw, select_fields)}

        for row in self._connection.execute(select_query):
            yield decode_row(row)

    def bulk_simple_dataset_search(self, products=None, batch_size=0):
        """
        Perform bulk database reads (e.g. for index cloning)

        Note that this operates with product ids to prevent an unnecessary join to the Product table.

        :param products: Optional iterable of product IDs.  Only fetch nominated products.
        :param batch_size: Number of streamed rows to fetch from database at once.
                           Defaults to zero, which means no streaming.
                           Note streaming is only supported inside a transaction.
        :return: Iterable of tuples of:
                 * Product ID
                 * Dataset metadata document
                 * array of uris
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        query = select(
            *_dataset_bulk_select_fields()
        ).select_from(Dataset).where(
            Dataset.archived.is_(None)
        )
        if products:
            query = query.where(Dataset.product_ref.in_(products))

        if batch_size > 0:
            conn = self._connection.execution_options(stream_results=True, yield_per=batch_size)
        else:
            conn = self._connection
        return conn.execute(query)

    def get_all_lineage(self, batch_size: int):
        """
        Stream all lineage data in bulk (e.g. for index cloning)

        :param batch_size: The number of lineage records to return at once.
        :return: Streamable SQLAlchemy result object.
        """
        if batch_size > 0 and not self.in_transaction:
            raise ValueError("Postgresql bulk reads must occur within a transaction.")
        query = select(
            DatasetLineage.derived_dataset_ref,
            DatasetLineage.classifier,
            DatasetLineage.source_dataset_ref
        )
        return self._connection.execution_options(stream_results=True, yield_per=batch_size).execute(query)

    def insert_lineage_bulk(self, values):
        """
        Insert bulk lineage records (e.g. for index cloning)

        :param values: An array of values dicts for bulk insert
        :return: Tuple[count of rows loaded, count of rows skipped]
        """
        requested = len(values)
        # Simple bulk insert with on_conflict_do_nothing.
        # No need to check referential integrity as this is an external lineage index driver.
        res = self._connection.execute(
            insert(DatasetLineage).on_conflict_do_nothing(),
            values
        )
        return res.rowcount, requested - res.rowcount

    def get_duplicates(self,
                       match_fields: Sequence[PgField],
                       expressions: Sequence[PgExpression]) -> Iterable[dict[str, Any]]:
        # TODO
        if "time" in [f.name for f in match_fields]:
            yield from self.get_duplicates_with_time(match_fields, expressions)

        group_expressions = tuple(f.alchemy_expression for f in match_fields)
        join_tables = PostgisDbAPI._join_tables(expressions, match_fields)

        query = select(
            func.array_agg(Dataset.id).label("ids"),
            *group_expressions
        ).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)

        query = query.where(
            and_(Dataset.archived.is_(None), *(PostgisDbAPI._alchemify_expressions(expressions)))
        ).group_by(
            *group_expressions
        ).having(
            func.count(Dataset.id) > 1
        )
        for row in self._connection.execute(query):
            drow = {"ids": row.ids}
            for f in match_fields:
                drow[f.name] = getattr(row, f.name)
            yield drow

    def get_duplicates_with_time(
            self, match_fields: Sequence[PgField], expressions: Sequence[PgExpression]
    ) -> Iterable[dict[str, Any]]:
        fields = []
        for fld in match_fields:
            if fld.name == "time":
                time_field = type_cast(DateRangeDocField, fld)
            else:
                fields.append(fld.alchemy_expression)

        join_tables = PostgisDbAPI._join_tables(expressions, match_fields)

        cols = [Dataset.id, time_field.expression_with_leniency.label('time'), *fields]
        query = select(
            *cols
        ).select_from(Dataset)
        for joins in join_tables:
            query = query.join(*joins)

        query = query.where(
            and_(Dataset.archived.is_(None), *(PostgisDbAPI._alchemify_expressions(expressions)))
        )

        t1 = query.alias("t1")
        t2 = query.alias("t2")

        time_overlap = select(
            t1.c.id,
            t1.c.time.intersection(t2.c.time).label('time_intersect'),
            *fields
        ).select_from(
            t1.join(
                t2,
                and_(t1.c.time.overlaps(t2.c.time), t1.c.id != t2.c.id)
            )
        )

        query = select(
            func.array_agg(func.distinct(time_overlap.c.id)).label("ids"),
            *fields,  # type: ignore[arg-type]
            text("time_intersect as time")
        ).select_from(
            time_overlap  # type: ignore[arg-type]
        ).group_by(
            *fields, text("time_intersect")
        ).having(
            func.count(time_overlap.c.id) > 1
        )

        for row in self._connection.execute(query):
            # TODO: Use decode_rows above - would require creating a field class for the ids array.
            drow: dict[str, Any] = {
                "ids": row.ids,
            }
            for f in fields:
                drow[f.key] = getattr(row, f.key)  # type: ignore[union-attr]
            drow["time"] = time_field.normalise_value((row.time.lower, row.time.upper))
            yield drow

    def count_datasets(self, expressions, archived: bool | None = False, geom: Geometry | None = None):
        """
        :type expressions: tuple[datacube.drivers.postgis._fields.PgExpression]
        :rtype: int
        """

        raw_expressions = self._alchemify_expressions(expressions)
        if archived:
            where_expressions = and_(Dataset.archived.is_not(None), *raw_expressions)
        elif archived is not None:
            where_expressions = and_(Dataset.archived.is_(None), *raw_expressions)
        else:
            where_expressions = and_(*raw_expressions)

        query = select(func.count(Dataset.id))
        if geom:
            SpatialIndex, spatialquery = self.geospatial_query(geom)
            where_expressions = and_(where_expressions, spatialquery)
            query = query.join(SpatialIndex)

        select_query = query.where(where_expressions)
        return self._connection.scalar(select_query)

    def count_datasets_through_time(self, start, end, period, time_field, expressions):
        """
        :type period: str
        :type start: datetime.datetime
        :type end: datetime.datetime
        :type expressions: tuple[datacube.drivers.postgis._fields.PgExpression]
        :rtype: list[((datetime.datetime, datetime.datetime), int)]
        """

        results = self._connection.execute(
            self.count_datasets_through_time_query(start, end, period, time_field, expressions)
        )

        for time_period, dataset_count in results:
            # if not time_period.upper_inf:
            yield Range(time_period.lower, time_period.upper), dataset_count

    def count_datasets_through_time_query(self, start, end, period, time_field, expressions):
        raw_expressions = self._alchemify_expressions(expressions)

        start_times = select(
            func.generate_series(start, end, cast(period, INTERVAL)).label('start_time'),
        ).alias('start_times')

        time_range_select = (
            select(
                func.tstzrange(
                    start_times.c.start_time,
                    func.lead(start_times.c.start_time).over()
                ).label('time_period'),
            )
        ).alias('all_time_ranges')

        # Exclude the trailing (end time to infinite) row. Is there a simpler way?
        time_ranges = (
            select(
                time_range_select,
            ).where(
                ~func.upper_inf(time_range_select.c.time_period)
            )
        ).alias('time_ranges')

        count_query = select(func.count('*'))
        join_tables = self._join_tables(expressions)
        for joins in join_tables:
            count_query = count_query.join(*joins)
        count_query = count_query.where(
            and_(
                time_field.alchemy_expression.overlaps(time_ranges.c.time_period),
                Dataset.archived == None,
                *raw_expressions
            )
        )

        return select(time_ranges.c.time_period, count_query.label('dataset_count'))

    def update_search_index(self, product_names: Sequence[str] = [], dsids: Sequence[DSID] = []):
        """
        Update search indexes
        :param product_names: Product names to update
        :param dsids: Dataset IDs to update

        if neither product_names nor dataset ids are supplied, update nothing (N.B. NOT all datasets)

        if both are supplied, both the named products and identified datasets are updated.

        :return:  Number of datasets whose search indexes have been updated.
        """
        if not product_names and not dsids:
            return 0

        ds_query = select(
            Dataset.id,
            Dataset.metadata_doc,
            MetadataType.definition,
        ).select_from(Dataset).join(MetadataType)
        if product_names:
            ds_query = ds_query.join(Product)
        if product_names and dsids:
            ds_query = ds_query.where(
                or_(
                    Product.name.in_(product_names),
                    Dataset.id.in_(dsids)
                )
            )
        elif product_names:
            ds_query = ds_query.where(
                Product.name.in_(product_names)
            )
        elif dsids:
            ds_query = ds_query.where(
                Dataset.id.in_(dsids)
            )
        rowcount = 0
        for result in self._connection.execute(ds_query):
            dsid, ds_metadata, mdt_def = result
            search_field_vals = extract_dataset_search_fields(ds_metadata, mdt_def)
            for field_name, field_info in search_field_vals.items():
                fld_type, fld_val = field_info
                search_idx = search_field_index_map[fld_type]
                self.insert_dataset_search(search_idx, dsid, field_name, fld_val)
            rowcount += 1
        return rowcount

    def update_spindex(self, crs_seq: Sequence[CRS] = [],
                       product_names: Sequence[str] = [],
                       dsids: Sequence[DSID] = []) -> int:
        """
        Update a spatial index
        :param crs: CRSs for Spatial Indexes to update. Default=all indexes
        :param product_names: Product names to update
        :param dsids: Dataset IDs to update

        if neither product_names nor dataset ids are supplied, update for all datasets.

        if both are supplied, both the named products and identified datasets are updated.

        :return:  Number of spatial index entries updated or verified as unindexed.
        """
        verified = 0
        if crs_seq:
            crses = [crs for crs in crs_seq]
        else:
            crses = self._db.spatially_indexed_crses()

        # Update implementation.
        # Design will change, but this method should be fairly low level to be as efficient as possible
        query = select(
            Dataset.id,
            Dataset.metadata_doc["grid_spatial"]["projection"]
        ).select_from(Dataset)
        if product_names:
            query = query.join(Product)
        if product_names and dsids:
            query = query.where(
                or_(
                    Product.name.in_(product_names),
                    Dataset.id.in_(dsids)
                )
            )
        elif product_names:
            query = query.where(
                Product.name.in_(product_names)
            )
        elif dsids:
            query = query.where(
                Dataset.id.in_(dsids)
            )

        def xytuple(o):
            return (o['x'], o['y'])

        for result in self._connection.execute(query):
            dsid = result[0]
            geom = extract_geometry_from_eo3_projection(result[1])
            if not geom:
                verified += 1
                continue
            for crs in crses:
                self.insert_dataset_spatial(dsid, crs, geom)
                verified += 1

        return verified

    @staticmethod
    def _join_tables(expressions=None, fields=None):
        join_args = set()
        if expressions:
            join_args.update(expression.field.dataset_join_args for expression in expressions)
        if fields:
            join_args.update(field.dataset_join_args for field in fields)
        join_args.discard((Dataset.__table__,))
        # Sort simple joins before qualified joins
        return sorted(join_args, key=len)

    def get_product(self, id_):
        return self._connection.execute(
            select(Product).where(Product.id == id_)
        ).first()

    def get_metadata_type(self, id_):
        return self._connection.execute(
            select(MetadataType).where(MetadataType.id == id_)
        ).first()

    def get_product_by_name(self, name):
        return self._connection.execute(
            select(Product).where(Product.name == name)
        ).first()

    def get_metadata_type_by_name(self, name):
        return self._connection.execute(
            select(MetadataType).where(MetadataType.name == name)
        ).first()

    def insert_product(self,
                       name,
                       metadata,
                       metadata_type_id,
                       definition):

        res = self._connection.execute(
            insert(Product).values(
                name=name,
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition
            )
        )

        type_id = res.inserted_primary_key[0]

        return type_id

    def insert_product_bulk(self, values):
        requested = len(values)
        res = self._connection.execute(insert(Product), values)
        return res.rowcount, requested - res.rowcount

    def update_product(self,
                       name,
                       metadata,
                       metadata_type_id,
                       definition,
                       update_metadata_type=False):
        res = self._connection.execute(
            update(Product).returning(Product.id).where(
                Product.name == name
            ).values(
                metadata=metadata,
                metadata_type_ref=metadata_type_id,
                definition=definition
            )
        )
        prod_id = res.first()[0]

        if update_metadata_type:
            if not self._connection.in_transaction():
                raise RuntimeError('Must update metadata types in transaction')

            self._connection.execute(
                update(Dataset).where(
                    Dataset.product_ref == prod_id
                ).values(
                    metadata_type_ref=metadata_type_id,
                )
            )

        return prod_id

    def delete_product(self, name):
        res = self._connection.execute(
            delete(Product).returning(Product.id).where(Product.name == name)
        )

        return res.first()[0]

    def insert_metadata_type(self, name, definition):
        res = self._connection.execute(
            insert(MetadataType).values(
                name=name,
                definition=definition
            )
        )
        return res.inserted_primary_key[0]

    def insert_metadata_bulk(self, values):
        requested = len(values)
        res = self._connection.execute(
            insert(MetadataType).on_conflict_do_nothing(index_elements=['id']),
            values
        )
        return res.rowcount, requested - res.rowcount

    def update_metadata_type(self, name, definition):
        res = self._connection.execute(
            update(MetadataType).returning(MetadataType.id).where(
                MetadataType.name == name
            ).values(
                name=name,
                definition=definition
            )
        )
        return res.first()[0]

    def _get_active_field_names(fields, metadata_doc):
        for field in fields.values():
            if field.can_extract:
                try:
                    value = field.extract(metadata_doc)
                    if value is not None:
                        yield field.name
                except (AttributeError, KeyError, ValueError):
                    continue

    def get_all_products(self):
        return self._connection.execute(
            select(Product).order_by(Product.name.asc())
        ).fetchall()

    def get_all_product_docs(self):
        return self._connection.execute(
            select(Product.definition)
        )

    def _get_products_for_metadata_type(self, id_):
        return self._connection.execute(
            select(Product).where(
                Product.metadata_type_ref == id_
            ).order_by(
                Product.name.asc()
            )).fetchall()

    def get_all_metadata_types(self):
        return self._connection.execute(select(MetadataType).order_by(MetadataType.name.asc())).fetchall()

    def get_all_metadata_type_defs(self):
        for r in self._connection.execute(select(MetadataType.definition).order_by(MetadataType.name.asc())):
            yield r[0]

    def get_location(self, dataset_id):
        return self._connection.execute(
            select(Dataset.uri).where(
                Dataset.id == dataset_id
            )
        ).first()

    def remove_location(self, dataset_id, uri):
        """
        Remove a dataset's location

        :returns bool: Was the location deleted?
        """
        scheme, body = split_uri(uri)
        res = self._connection.execute(
            update(Dataset).where(
                and_(
                    Dataset.id == dataset_id,
                    Dataset.uri_scheme == scheme,
                    Dataset.uri_body == body
                )
            ).values(
                uri_scheme=None,
                uri_body=None
            )
        )
        return res.rowcount > 0

    def __repr__(self):
        return "PostgresDb<connection={!r}>".format(self._connection)

    def list_users(self):
        result = self._connection.execute(text("""
            select
                group_role.rolname as role_name,
                user_role.rolname as user_name,
                pg_catalog.shobj_description(user_role.oid, 'pg_authid') as description
            from pg_roles group_role
            inner join pg_auth_members am on am.roleid = group_role.oid
            inner join pg_roles user_role on am.member = user_role.oid
            where (group_role.rolname like 'odc_%%') and not (user_role.rolname like 'odc_%%')
            order by group_role.oid asc, user_role.oid asc;
        """))
        for row in result:
            yield _core.from_pg_role(row.role_name), row.user_name, row.description

    def create_user(self, username, password, role, description=None):
        pg_role = _core.to_pg_role(role)
        username = escape_pg_identifier(self._connection, username)
        sql = text('create user {username} password :password in role {role}'.format(username=username, role=pg_role))
        self._connection.execute(sql, {"password": password})
        if description:
            sql = text('comment on role {username} is :description'.format(username=username))
            self._connection.execute(sql, {"description": description})

    def drop_users(self, users: Iterable[str]) -> None:
        for username in users:
            sql = text('drop role {username}'.format(username=escape_pg_identifier(self._connection, username)))
            self._connection.execute(sql)

    def grant_role(self, role: str, users: Iterable[str]) -> None:
        """
        Grant a role to a user.
        """
        pg_role = _core.to_pg_role(role)

        for user in users:
            if not _core.has_role(self._connection, user):
                raise ValueError('Unknown user %r' % user)

        _core.grant_role(self._connection, pg_role, users)

    def insert_home(self, home: str, ids: Iterable[uuid.UUID], allow_updates: bool) -> int:
        """
        Set home for multiple IDs (but one home value)

        :param home: The home value to set
        :param ids: The IDs to set it for
        :param allow_updates: If False only inserts are allowed
        :return: number of database records updated or added.
        """
        values = [
            {"dataset_ref": id_, "home": home}
            for id_ in ids
        ]
        qry = insert(DatasetHome)
        if allow_updates:
            qry = qry.on_conflict_do_update(
                index_elements=["dataset_ref"],
                set_={"home": home},
                where=(DatasetHome.home != home))
        try:
            res = self._connection.execute(
                qry,
                values
            )
            return res.rowcount
        except IntegrityError:
            return 0

    def delete_home(self, ids):
        """
        Delete the home value for the specified IDs

        :param ids: The IDs to delete home for
        :return: The number of hone records deleted from the databes
        """
        res = self._connection.execute(
            delete(DatasetHome).where(DatasetHome.dataset_ref.in_(ids))
        )
        return res.rowcount

    def select_homes(self, ids):
        """
        Find homes for IDs.

        :param ids: Iterable of IDs
        :return: Mapping of ID to home string for IDs found in database.
        """
        results = self._connection.execute(
            select(DatasetHome).where(DatasetHome.dataset_ref.in_(ids))
        )
        return {
            row.dataset_ref: row.home
            for row in results
        }

    def get_all_relations(self, dsids: Iterable[uuid.UUID]) -> Iterable[LineageRelation]:
        """
        Fetch all lineage relations in the database involving a set on dataset IDs.

        :param dsids: Iterable of dataset IDs
        :return: Iterable of LineageRelation objects.
        """
        results = self._connection.execute(
            select(DatasetLineage).where(or_(
                DatasetLineage.derived_dataset_ref.in_(dsids),
                DatasetLineage.source_dataset_ref.in_(dsids)
            ))
        )
        for rel in results:
            yield LineageRelation(classifier=rel.classifier,
                                  source_id=rel.source_dataset_ref,
                                  derived_id=rel.derived_dataset_ref)

    def write_relations(self, relations: Iterable[LineageRelation], allow_updates: bool) -> int:
        """
        Write a set of LineageRelation objects to the database.

        :param relations: An Iterable of LineageRelation objects
        :param allow_updates: if False, only allow adding new relations, not updating old ones.
        :return: Count of database rows affected
        """
        affected = 0
        if allow_updates:
            by_classifier: dict[str, Any] = {}
            for rel in relations:
                db_repr = {
                    "derived_dataset_ref": rel.derived_id,
                    "source_dataset_ref": rel.source_id,
                    "classifier": rel.classifier
                }
                if rel.classifier in by_classifier:
                    by_classifier[rel.classifier].append(db_repr)
                else:
                    by_classifier[rel.classifier] = [db_repr]
                for classifier, values in by_classifier.items():
                    qry = insert(DatasetLineage).on_conflict_do_update(
                        index_elements=["derived_dataset_ref", "source_dataset_ref"],
                        set_={"classifier": classifier},
                        where=(DatasetLineage.classifier != classifier))
                    res = self._connection.execute(qry, values)
                    affected += res.rowcount
        else:
            for rel in relations:
                values = [
                    {
                        "derived_dataset_ref": rel.derived_id,
                        "source_dataset_ref": rel.source_id,
                        "classifier": rel.classifier
                    }
                ]
                qry = insert(DatasetLineage)
                try:
                    res = self._connection.execute(
                        qry, values
                    )
                    affected += res.rowcount
                except IntegrityError:
                    return 0
        return affected

    def load_lineage_relations(self,
                               roots: Iterable[uuid.UUID],
                               direction: LineageDirection,
                               depth: int,
                               ids_so_far: Optional[Set[uuid.UUID]] = None) -> Iterable[LineageRelation]:
        """
        Read from the database all indexed LineageRelation objects required to build all LineageTrees with
        the given roots, direction and depth.

        :param roots: Iterable of root dataset ids
        :param direction: tree direction
        :param depth: Maximum tree depth - zero indicates unlimited depth.
        :param ids_so_far: Used for maintaining state through recursion - expected to be None on initial call
        :return: Iterable of LineageRelation objects read from database
        """
        # Naive manually-recursive initial implementation.
        # TODO: Reimplement using WITH RECURSIVE query
        if ids_so_far is None:
            ids_so_far = set(roots)
        qry = select(DatasetLineage)
        if direction == LineageDirection.SOURCES:
            qry = qry.where(DatasetLineage.derived_dataset_ref.in_(roots))
        else:
            qry = qry.where(DatasetLineage.source_dataset_ref.in_(roots))
        relations = []
        next_lvl_ids = set()
        results = self._connection.execute(qry)
        for row in results:
            rel = LineageRelation(classifier=row.classifier,
                                  source_id=row.source_dataset_ref,
                                  derived_id=row.derived_dataset_ref)
            relations.append(rel)
            if direction == LineageDirection.SOURCES:
                next_id = rel.source_id
            else:
                next_id = rel.derived_id
            if next_id not in ids_so_far:
                next_lvl_ids.add(next_id)
                ids_so_far.add(next_id)
        next_depth = depth - 1
        recurse = True
        if depth == 0:
            next_depth = 0
        elif depth == 1:
            recurse = False
        if recurse and next_lvl_ids:
            relations.extend(self.load_lineage_relations(next_lvl_ids, direction, next_depth, ids_so_far))
        return relations

    def remove_lineage_relations(self,
                                 ids: Iterable[DSID],
                                 direction: LineageDirection) -> int:
        """
        Remove lineage relations from the provided ids in the specified direction.

        Note no depth parameter - depth is effectively always 1.

        :param ids: Iterable of IDs to remove lineage information for.
        :param direction: Remove the source or derived lineage relation records
        :return: Return number of relation records deleted.
        """
        qry = delete(DatasetLineage)
        if direction == LineageDirection.SOURCES:
            qry = qry.where(DatasetLineage.derived_dataset_ref.in_(ids))
        else:
            qry = qry.where(DatasetLineage.source_dataset_ref.in_(ids))
        results = self._connection.execute(qry)
        return results.rowcount

    def temporal_extent_by_prod(self, product_id: int) -> tuple[datetime.datetime, datetime.datetime]:
        query = self.temporal_extent_full().where(Dataset.product_ref == product_id)
        res = self._connection.execute(query)
        for tmin, tmax in res:
            return (time_min.normalise_value(tmin), time_max.normalise_value(tmax))
        raise RuntimeError("Product has no datasets and therefore no temporal extent")

    def temporal_extent_by_ids(self, ids: Iterable[DSID]) -> tuple[datetime.datetime, datetime.datetime]:
        query = self.temporal_extent_full().where(Dataset.id.in_(ids))
        res = self._connection.execute(query)
        for tmin, tmax in res:
            return (time_min.normalise_value(tmin), time_max.normalise_value(tmax))
        raise ValueError("no dataset ids provided")

    def temporal_extent_full(self) -> Select:
        # Hardcode eo3 standard time locations - do not use this approach in a legacy index driver.

        return select(
            func.min(time_min.alchemy_expression), func.max(time_max.alchemy_expression)
        )

    def find_most_recent_change(self, product_id: int):
        """
        Find the database-local time of the last dataset that changed for this product.
        """
        return self._connection.execute(
            select(
                func.max(
                    func.greatest(
                        Dataset.added,
                        column("updated"),
                    )
                )
            ).where(Dataset.product_ref == product_id)
        ).scalar()



# This file is part of the Open Data Cube, see https://opendatacube.org for more information
#
# Copyright (c) 2015-2024 ODC Contributors
# SPDX-License-Identifier: Apache-2.0


from .exceptions import ConfigException
from .utils import ConfigDict, check_valid_env_name, check_valid_option, smells_like_ini
from .cfg import find_config, CfgFormat, parse_text
from .opt import ODCOptionHandler, IntOptionHandler, IAMAuthenticationOptionHandler, PostgresURLOptionHandler
from .opt import config_options_for_psql_driver, psql_url_from_config
from .api import GeneralisedPath, GeneralisedCfg, GeneralisedEnv, GeneralisedRawCfg, ODCConfig, ODCEnvironment


__all__ = [
    "ConfigException", "ConfigDict",
    "ODCConfig", "ODCEnvironment",
    "find_config", "CfgFormat", "parse_text",
    "GeneralisedPath", "GeneralisedCfg", "GeneralisedEnv", "GeneralisedRawCfg",
    "ODCConfig", "ODCEnvironment",
    "ODCOptionHandler", "IntOptionHandler", "IAMAuthenticationOptionHandler", "PostgresURLOptionHandler",
    "config_options_for_psql_driver", "psql_url_from_config",
    "check_valid_env_name", "check_valid_option", "smells_like_ini"
]
